// SPDX-License-Identifier: GPL-2.0
/*
 * MediaTek clock driver for MT8188 SoC
 *
 * Copyright (C) 2024 BayLibre, SAS
 * Copyright (c) 2024 MediaTek Inc.
 * Authors: Julien Masson <jmasson@baylibre.com>
 *          Garmin Chang <garmin.chang@mediatek.com>
 */

#include <common.h>
#include <clk-uclass.h>
#include <dm/device_compat.h>
#include <dm.h>
#include <asm/io.h>
#include <dt-bindings/clock/mediatek,mt8188-clk.h>

#include "clk-mtk.h"

#define MT8188_PLL_FMAX		(3800UL * MHZ)
#define MT8188_PLL_FMIN		(1500UL * MHZ)

/* Missing topckgen clocks definition in dt-bindings */
#define CLK_TOP_ADSPPLL		206
#define CLK_TOP_CLK13M		207
#define CLK_TOP_CLK26M		208
#define CLK_TOP_CLK32K		209
#define CLK_TOP_IMGPLL		210
#define CLK_TOP_MSDCPLL		211
#define CLK_TOP_ULPOSC1_CK1	212
#define CLK_TOP_ULPOSC_CK1	213

/* apmixedsys */
#define PLL(_id, _reg, _pwr_reg, _en_mask, _flags, _pcwbits, _pd_reg,	\
	    _pd_shift, _pcw_reg, _pcw_shift) {				\
		.id = _id,						\
		.reg = _reg,						\
		.pwr_reg = _pwr_reg,					\
		.en_mask = _en_mask,					\
		.rst_bar_mask = BIT(23),				\
		.fmin = MT8188_PLL_FMIN,				\
		.fmax = MT8188_PLL_FMAX,				\
		.flags = _flags,					\
		.pcwbits = _pcwbits,					\
		.pcwibits = 8,						\
		.pd_reg = _pd_reg,					\
		.pd_shift = _pd_shift,					\
		.pcw_reg = _pcw_reg,					\
		.pcw_shift = _pcw_shift,				\
	}

static const struct mtk_pll_data apmixed_plls[] = {
	PLL(CLK_APMIXED_ETHPLL, 0x044C, 0x0458, 0, 0, 22, 0x0450, 24,
	    0x0450, 0),
	PLL(CLK_APMIXED_MSDCPLL, 0x0514, 0x0520, 0, 0, 22, 0x0518, 24,
	    0x0518, 0),
	PLL(CLK_APMIXED_TVDPLL1, 0x0524, 0x0530, 0, 0, 22, 0x0528, 24,
	    0x0528, 0),
	PLL(CLK_APMIXED_TVDPLL2, 0x0534, 0x0540, 0, 0, 22, 0x0538, 24,
	    0x0538, 0),
	PLL(CLK_APMIXED_MMPLL, 0x0544, 0x0550, 0xff000000, HAVE_RST_BAR,
	    22, 0x0548, 24, 0x0548, 0),
	PLL(CLK_APMIXED_MAINPLL, 0x045C, 0x0468, 0xff000000, HAVE_RST_BAR,
	    22, 0x0460, 24, 0x0460, 0),
	PLL(CLK_APMIXED_IMGPLL, 0x0554, 0x0560, 0, 0, 22, 0x0558, 24,
	    0x0558, 0),
	PLL(CLK_APMIXED_UNIVPLL, 0x0504, 0x0510, 0xff000000, HAVE_RST_BAR,
	    22, 0x0508, 24, 0x0508, 0),
	PLL(CLK_APMIXED_ADSPPLL, 0x042C, 0x0438, 0, 0, 22, 0x0430, 24,
	    0x0430, 0),
	PLL(CLK_APMIXED_APLL1, 0x0304, 0x0314, 0, 0, 32, 0x0308, 24,
	    0x030C, 0),
	PLL(CLK_APMIXED_APLL2, 0x0318, 0x0328, 0, 0, 32, 0x031C, 24,
	    0x0320, 0),
	PLL(CLK_APMIXED_APLL3, 0x032C, 0x033C, 0, 0, 32, 0x0330, 24,
	    0x0334, 0),
	PLL(CLK_APMIXED_APLL4, 0x0404, 0x0414, 0, 0, 32, 0x0408, 24,
	    0x040C, 0),
	PLL(CLK_APMIXED_APLL5, 0x0418, 0x0428, 0, 0, 32, 0x041C, 24,
	    0x0420, 0),
	PLL(CLK_APMIXED_MFGPLL, 0x0340, 0x034C, 0, 0, 22, 0x0344, 24,
	    0x0344, 0),
};

static const struct mtk_fixed_clk top_fixed_clks[] = {
	FIXED_CLK(CLK_TOP_ULPOSC1, CLK_XTAL, 260000000),
	FIXED_CLK(CLK_TOP_MPHONE_SLAVE_BCK, CLK_XTAL, 49152000),
	FIXED_CLK(CLK_TOP_PAD_FPC, CLK_XTAL, 50000000),
	FIXED_CLK(CLK_TOP_466M_FMEM, CLK_XTAL, 533000000),
	FIXED_CLK(CLK_TOP_PEXTP_PIPE, CLK_XTAL, 250000000),
	FIXED_CLK(CLK_TOP_DSI_PHY, CLK_XTAL, 500000000),
	FIXED_CLK(CLK_TOP_CLK26M, CLK_XTAL, 260000000),
	FIXED_CLK(CLK_TOP_CLK32K, CLK_XTAL, 32000),
};

static const struct mtk_fixed_clk * find_top_fixed_clk(unsigned long id)
{
	const struct mtk_fixed_clk *fclk = NULL;
	u32 i;

	for (i = 0; i < ARRAY_SIZE(top_fixed_clks); i++) {
		if (id == top_fixed_clks[i].id) {
			fclk = &top_fixed_clks[i];
			break;
		}
	}

	return fclk;
}

#define FACTOR0(_id, _parent, _mult, _div)	\
	FACTOR(_id, _parent, _mult, _div, CLK_PARENT_APMIXED)

#define FACTOR1(_id, _parent, _mult, _div)	\
	FACTOR(_id, _parent, _mult, _div, CLK_PARENT_TOPCKGEN)

static const struct mtk_fixed_factor top_fixed_divs[] = {
	FACTOR0(CLK_TOP_MAINPLL_D3, CLK_APMIXED_MAINPLL, 1, 3),
	FACTOR0(CLK_TOP_MAINPLL_D4, CLK_APMIXED_MAINPLL, 1, 4),
	FACTOR1(CLK_TOP_MAINPLL_D4_D2, CLK_TOP_MAINPLL_D4, 1, 2),
	FACTOR1(CLK_TOP_MAINPLL_D4_D4, CLK_TOP_MAINPLL_D4, 1, 4),
	FACTOR1(CLK_TOP_MAINPLL_D4_D8, CLK_TOP_MAINPLL_D4, 1, 8),
	FACTOR0(CLK_TOP_MAINPLL_D5, CLK_APMIXED_MAINPLL, 1, 5),
	FACTOR1(CLK_TOP_MAINPLL_D5_D2, CLK_TOP_MAINPLL_D5, 1, 2),
	FACTOR1(CLK_TOP_MAINPLL_D5_D4, CLK_TOP_MAINPLL_D5, 1, 4),
	FACTOR1(CLK_TOP_MAINPLL_D5_D8, CLK_TOP_MAINPLL_D5, 1, 8),
	FACTOR0(CLK_TOP_MAINPLL_D6, CLK_APMIXED_MAINPLL, 1, 6),
	FACTOR1(CLK_TOP_MAINPLL_D6_D2, CLK_TOP_MAINPLL_D6, 1, 2),
	FACTOR1(CLK_TOP_MAINPLL_D6_D4, CLK_TOP_MAINPLL_D6, 1, 4),
	FACTOR1(CLK_TOP_MAINPLL_D6_D8, CLK_TOP_MAINPLL_D6, 1, 8),
	FACTOR0(CLK_TOP_MAINPLL_D7, CLK_APMIXED_MAINPLL, 1, 7),
	FACTOR1(CLK_TOP_MAINPLL_D7_D2, CLK_TOP_MAINPLL_D7, 1, 2),
	FACTOR1(CLK_TOP_MAINPLL_D7_D4, CLK_TOP_MAINPLL_D7, 1, 4),
	FACTOR1(CLK_TOP_MAINPLL_D7_D8, CLK_TOP_MAINPLL_D7, 1, 8),
	FACTOR0(CLK_TOP_MAINPLL_D9, CLK_APMIXED_MAINPLL, 1, 9),
	FACTOR0(CLK_TOP_UNIVPLL_D2, CLK_APMIXED_UNIVPLL, 1, 2),
	FACTOR0(CLK_TOP_UNIVPLL_D3, CLK_APMIXED_UNIVPLL, 1, 3),
	FACTOR0(CLK_TOP_UNIVPLL_D4, CLK_APMIXED_UNIVPLL, 1, 4),
	FACTOR1(CLK_TOP_UNIVPLL_D4_D2, CLK_TOP_UNIVPLL_D4, 1, 2),
	FACTOR1(CLK_TOP_UNIVPLL_D4_D4, CLK_TOP_UNIVPLL_D4, 1, 4),
	FACTOR1(CLK_TOP_UNIVPLL_D4_D8, CLK_TOP_UNIVPLL_D4, 1, 8),
	FACTOR0(CLK_TOP_UNIVPLL_D5, CLK_APMIXED_UNIVPLL, 1, 5),
	FACTOR1(CLK_TOP_UNIVPLL_D5_D2, CLK_TOP_UNIVPLL_D5, 1, 2),
	FACTOR1(CLK_TOP_UNIVPLL_D5_D4, CLK_TOP_UNIVPLL_D5, 1, 4),
	FACTOR1(CLK_TOP_UNIVPLL_D5_D8, CLK_TOP_UNIVPLL_D5, 1, 8),
	FACTOR0(CLK_TOP_UNIVPLL_D6, CLK_APMIXED_UNIVPLL, 1, 6),
	FACTOR1(CLK_TOP_UNIVPLL_D6_D2, CLK_TOP_UNIVPLL_D6, 1, 2),
	FACTOR1(CLK_TOP_UNIVPLL_D6_D4, CLK_TOP_UNIVPLL_D6, 1, 4),
	FACTOR1(CLK_TOP_UNIVPLL_D6_D8, CLK_TOP_UNIVPLL_D6, 1, 8),
	FACTOR0(CLK_TOP_UNIVPLL_D7, CLK_APMIXED_UNIVPLL, 1, 7),
	FACTOR0(CLK_TOP_UNIVPLL_192M, CLK_APMIXED_UNIVPLL, 1, 13),
	FACTOR1(CLK_TOP_UNIVPLL_192M_D4, CLK_TOP_UNIVPLL_192M, 1, 4),
	FACTOR1(CLK_TOP_UNIVPLL_192M_D8, CLK_TOP_UNIVPLL_192M, 1, 8),
	FACTOR1(CLK_TOP_UNIVPLL_192M_D10, CLK_TOP_UNIVPLL_192M, 1, 10),
	FACTOR1(CLK_TOP_UNIVPLL_192M_D16, CLK_TOP_UNIVPLL_192M, 1, 16),
	FACTOR1(CLK_TOP_UNIVPLL_192M_D32, CLK_TOP_UNIVPLL_192M, 1, 32),
	FACTOR1(CLK_TOP_APLL1_D3, CLK_TOP_APLL1, 1, 3),
	FACTOR1(CLK_TOP_APLL1_D4, CLK_TOP_APLL1, 1, 4),
	FACTOR1(CLK_TOP_APLL2_D3, CLK_TOP_APLL2, 1, 3),
	FACTOR1(CLK_TOP_APLL2_D4, CLK_TOP_APLL2, 1, 4),
	FACTOR1(CLK_TOP_APLL3_D4, CLK_TOP_APLL3, 1, 4),
	FACTOR1(CLK_TOP_APLL4_D4, CLK_TOP_APLL4, 1, 4),
	FACTOR1(CLK_TOP_APLL5_D4, CLK_TOP_APLL5, 1, 4),
	FACTOR0(CLK_TOP_MMPLL_D4, CLK_APMIXED_MMPLL, 1, 4),
	FACTOR1(CLK_TOP_MMPLL_D4_D2, CLK_TOP_MMPLL_D4, 1, 2),
	FACTOR0(CLK_TOP_MMPLL_D5, CLK_APMIXED_MMPLL, 1, 5),
	FACTOR1(CLK_TOP_MMPLL_D5_D2, CLK_TOP_MMPLL_D5, 1, 2),
	FACTOR1(CLK_TOP_MMPLL_D5_D4, CLK_TOP_MMPLL_D5, 1, 4),
	FACTOR0(CLK_TOP_MMPLL_D6, CLK_APMIXED_MMPLL, 1, 6),
	FACTOR1(CLK_TOP_MMPLL_D6_D2, CLK_TOP_MMPLL_D6, 1, 2),
	FACTOR0(CLK_TOP_MMPLL_D7, CLK_APMIXED_MMPLL, 1, 7),
	FACTOR0(CLK_TOP_MMPLL_D9, CLK_APMIXED_MMPLL, 1, 9),
	FACTOR1(CLK_TOP_TVDPLL1_D2, CLK_TOP_TVDPLL1, 1, 2),
	FACTOR1(CLK_TOP_TVDPLL1_D4, CLK_TOP_TVDPLL1, 1, 4),
	FACTOR1(CLK_TOP_TVDPLL1_D8, CLK_TOP_TVDPLL1, 1, 8),
	FACTOR1(CLK_TOP_TVDPLL1_D16,  CLK_TOP_TVDPLL1, 1, 16),
	FACTOR1(CLK_TOP_TVDPLL2_D2, CLK_TOP_TVDPLL2, 1, 2),
	FACTOR1(CLK_TOP_TVDPLL2_D4, CLK_TOP_TVDPLL2, 1, 4),
	FACTOR1(CLK_TOP_TVDPLL2_D8, CLK_TOP_TVDPLL2, 1, 8),
	FACTOR1(CLK_TOP_TVDPLL2_D16, CLK_TOP_TVDPLL2, 1, 16),
	FACTOR0(CLK_TOP_MSDCPLL, CLK_APMIXED_MSDCPLL, 1, 1),
	FACTOR0(CLK_TOP_MSDCPLL_D2, CLK_APMIXED_MSDCPLL, 1, 2),
	FACTOR0(CLK_TOP_MSDCPLL_D16, CLK_APMIXED_MSDCPLL, 1, 16),
	FACTOR1(CLK_TOP_ETHPLL_D2, CLK_TOP_ETHPLL, 1, 2),
	FACTOR1(CLK_TOP_ETHPLL_D4, CLK_TOP_ETHPLL, 1, 4),
	FACTOR1(CLK_TOP_ETHPLL_D8, CLK_TOP_ETHPLL, 1, 8),
	FACTOR1(CLK_TOP_ETHPLL_D10, CLK_TOP_ETHPLL, 1, 10),
	FACTOR0(CLK_TOP_ADSPPLL_D2, CLK_APMIXED_ADSPPLL, 1, 2),
	FACTOR0(CLK_TOP_ADSPPLL_D4, CLK_APMIXED_ADSPPLL, 1, 4),
	FACTOR0(CLK_TOP_ADSPPLL_D8, CLK_APMIXED_ADSPPLL, 1, 8),
	FACTOR1(CLK_TOP_ULPOSC1_D2, CLK_TOP_ULPOSC1, 1, 2),
	FACTOR1(CLK_TOP_ULPOSC1_D4, CLK_TOP_ULPOSC1, 1, 4),
	FACTOR1(CLK_TOP_ULPOSC1_D8, CLK_TOP_ULPOSC1, 1, 8),
	FACTOR1(CLK_TOP_ULPOSC1_D7, CLK_TOP_ULPOSC1, 1, 7),
	FACTOR1(CLK_TOP_ULPOSC1_D10, CLK_TOP_ULPOSC1, 1, 10),
	FACTOR1(CLK_TOP_ULPOSC1_D16, CLK_TOP_ULPOSC1, 1, 16),
};

static const struct mtk_fixed_factor * find_top_fixed_div(unsigned long id)
{
	const struct mtk_fixed_factor *fdiv = NULL;
	u32 i;

	for (i = 0; i < ARRAY_SIZE(top_fixed_divs); i++) {
		if (id == top_fixed_divs[i].id) {
			fdiv = &top_fixed_divs[i];
			break;
		}
	}

	return fdiv;
}

static const int axi_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MAINPLL_D4_D4,
	CLK_TOP_MAINPLL_D7_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_MAINPLL_D5_D2,
	CLK_TOP_MAINPLL_D6_D2,
	CLK_TOP_ULPOSC1_D4
};

static const int spm_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_ULPOSC1_D10,
	CLK_TOP_MAINPLL_D7_D4,
	CLK_TOP_CLK32K
};

static const int scp_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_MAINPLL_D6,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_UNIVPLL_D3,
	CLK_TOP_MAINPLL_D3
};

static const int bus_aximem_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MAINPLL_D7_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_MAINPLL_D5_D2,
	CLK_TOP_MAINPLL_D6
};

static const int vpp_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_MAINPLL_D5_D2,
	CLK_TOP_MMPLL_D6_D2,
	CLK_TOP_UNIVPLL_D5_D2,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_MMPLL_D4_D2,
	CLK_TOP_MMPLL_D7,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_MAINPLL_D4,
	CLK_TOP_MMPLL_D5,
	CLK_TOP_TVDPLL1,
	CLK_TOP_TVDPLL2,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_MMPLL_D4
};

static const int ethdr_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_MAINPLL_D5_D2,
	CLK_TOP_MMPLL_D6_D2,
	CLK_TOP_UNIVPLL_D5_D2,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_MMPLL_D4_D2,
	CLK_TOP_MMPLL_D7,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_MAINPLL_D4,
	CLK_TOP_MMPLL_D5_D4,
	CLK_TOP_TVDPLL1,
	CLK_TOP_TVDPLL2,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_MMPLL_D4
};

static const int ipe_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_IMGPLL,
	CLK_TOP_MAINPLL_D4,
	CLK_TOP_MMPLL_D6,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_MAINPLL_D6,
	CLK_TOP_MMPLL_D4_D2,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_MMPLL_D6_D2,
	CLK_TOP_UNIVPLL_D5_D2,
	CLK_TOP_MAINPLL_D7
};

static const int cam_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_TVDPLL1,
	CLK_TOP_MAINPLL_D4,
	CLK_TOP_MMPLL_D4,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_UNIVPLL_D5,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_MMPLL_D7,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_IMGPLL
};

static const int ccu_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_MAINPLL_D4,
	CLK_TOP_UNIVPLL_D5,
	CLK_TOP_MAINPLL_D6,
	CLK_TOP_MMPLL_D6,
	CLK_TOP_MMPLL_D7,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_UNIVPLL_D7
};

static const int ccu_ahb_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_MAINPLL_D4,
	CLK_TOP_UNIVPLL_D5,
	CLK_TOP_MAINPLL_D6,
	CLK_TOP_MMPLL_D6,
	CLK_TOP_MMPLL_D7,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_UNIVPLL_D7
};

static const int img_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_IMGPLL,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_MAINPLL_D4,
	CLK_TOP_UNIVPLL_D5,
	CLK_TOP_MMPLL_D6,
	CLK_TOP_MMPLL_D7,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_MAINPLL_D6,
	CLK_TOP_MMPLL_D4_D2,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_UNIVPLL_D5_D2
};

static const int camtm_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D4_D4,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_UNIVPLL_D6_D4
};

static const int dsp_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_UNIVPLL_D5,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_MMPLL_D4,
	CLK_TOP_MAINPLL_D3,
	CLK_TOP_UNIVPLL_D3
};

static const int dsp1_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_UNIVPLL_D5,
	CLK_TOP_MMPLL_D5,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_MAINPLL_D3,
	CLK_TOP_UNIVPLL_D3
};

static const int dsp2_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_UNIVPLL_D5,
	CLK_TOP_MMPLL_D5,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_MAINPLL_D3,
	CLK_TOP_UNIVPLL_D3
};

static const int dsp3_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_UNIVPLL_D5,
	CLK_TOP_MMPLL_D5,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_MAINPLL_D3,
	CLK_TOP_UNIVPLL_D3
};

static const int dsp4_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_MAINPLL_D4,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_MMPLL_D4,
	CLK_TOP_MAINPLL_D3,
	CLK_TOP_UNIVPLL_D3
};

static const int dsp5_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_MAINPLL_D4,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_MMPLL_D4,
	CLK_TOP_MAINPLL_D3,
	CLK_TOP_UNIVPLL_D3
};

static const int dsp6_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_MAINPLL_D4,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_MMPLL_D4,
	CLK_TOP_MAINPLL_D3,
	CLK_TOP_UNIVPLL_D3
};

static const int dsp7_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_UNIVPLL_D5,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_MMPLL_D4,
	CLK_TOP_MAINPLL_D3,
	CLK_TOP_UNIVPLL_D3
};

static const int mfg_core_tmp_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MAINPLL_D5_D2,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_UNIVPLL_D7
};

static const int camtg_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_192M_D8,
	CLK_TOP_UNIVPLL_D6_D8,
	CLK_TOP_UNIVPLL_192M_D4,
	CLK_TOP_UNIVPLL_192M_D10,
	CLK_TOP_CLK13M,
	CLK_TOP_UNIVPLL_192M_D16,
	CLK_TOP_UNIVPLL_192M_D32
};

static const int camtg2_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_192M_D8,
	CLK_TOP_UNIVPLL_D6_D8,
	CLK_TOP_UNIVPLL_192M_D4,
	CLK_TOP_UNIVPLL_192M_D10,
	CLK_TOP_CLK13M,
	CLK_TOP_UNIVPLL_192M_D16,
	CLK_TOP_UNIVPLL_192M_D32
};

static const int camtg3_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_192M_D8,
	CLK_TOP_UNIVPLL_D6_D8,
	CLK_TOP_UNIVPLL_192M_D4,
	CLK_TOP_UNIVPLL_192M_D10,
	CLK_TOP_CLK13M,
	CLK_TOP_UNIVPLL_192M_D16,
	CLK_TOP_UNIVPLL_192M_D32
};

static const int uart_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D8
};

static const int spi_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MAINPLL_D5_D4,
	CLK_TOP_MAINPLL_D6_D4,
	CLK_TOP_UNIVPLL_D6_D4,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_MAINPLL_D6_D2,
	CLK_TOP_MAINPLL_D4_D4,
	CLK_TOP_UNIVPLL_D5_D4
};

static const int msdc5hclk_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_MAINPLL_D6_D2
};

static const int msdc50_0_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MSDCPLL,
	CLK_TOP_MSDCPLL_D2,
	CLK_TOP_UNIVPLL_D4_D4,
	CLK_TOP_MAINPLL_D6_D2,
	CLK_TOP_UNIVPLL_D4_D2
};

static const int msdc30_1_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_MAINPLL_D6_D2,
	CLK_TOP_MAINPLL_D7_D2,
	CLK_TOP_MSDCPLL_D2
};

static const int msdc30_2_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_MAINPLL_D6_D2,
	CLK_TOP_MAINPLL_D7_D2,
	CLK_TOP_MSDCPLL_D2
};

static const int intdir_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_MAINPLL_D4,
	CLK_TOP_UNIVPLL_D4
};

static const int aud_intbus_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MAINPLL_D4_D4,
	CLK_TOP_MAINPLL_D7_D4
};

static const int audio_h_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D7,
	CLK_TOP_APLL1,
	CLK_TOP_APLL2
};

static const int pwrap_ulposc_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_ULPOSC1_D10,
	CLK_TOP_ULPOSC1_D7,
	CLK_TOP_ULPOSC1_D8,
	CLK_TOP_ULPOSC1_D16,
	CLK_TOP_MAINPLL_D4_D8,
	CLK_TOP_UNIVPLL_D5_D8,
	CLK_TOP_TVDPLL1_D16
};

static const int atb_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_MAINPLL_D5_D2
};

static const int sspm_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MAINPLL_D7_D2,
	CLK_TOP_MAINPLL_D6_D2,
	CLK_TOP_MAINPLL_D5_D2,
	CLK_TOP_MAINPLL_D9,
	CLK_TOP_MAINPLL_D4_D2
};

/*
 * Both DP/eDP can be parented to TVDPLL1 and TVDPLL2, but we force using
 * TVDPLL1 on eDP and TVDPLL2 on DP to avoid changing the "other" PLL rate
 * in dual output case, which would lead to corruption of functionality loss.
 */
static const int dp_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_TVDPLL2_D2,
	CLK_TOP_TVDPLL2_D4,
	CLK_TOP_TVDPLL2_D8,
	CLK_TOP_TVDPLL2_D16
};

static const int edp_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_TVDPLL1_D2,
	CLK_TOP_TVDPLL1_D4,
	CLK_TOP_TVDPLL1_D8,
	CLK_TOP_TVDPLL1_D16
};

static const int dpi_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_TVDPLL1_D2,
	CLK_TOP_TVDPLL2_D2,
	CLK_TOP_TVDPLL1_D4,
	CLK_TOP_TVDPLL2_D4,
	CLK_TOP_TVDPLL1_D8,
	CLK_TOP_TVDPLL2_D8,
	CLK_TOP_TVDPLL1_D16,
	CLK_TOP_TVDPLL2_D16
};

static const int disp_pwm0_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D4,
	CLK_TOP_ULPOSC1_D2,
	CLK_TOP_ULPOSC1_D4,
	CLK_TOP_ULPOSC1_D16,
	CLK_TOP_ETHPLL_D4
};

static const int disp_pwm1_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D4,
	CLK_TOP_ULPOSC1_D2,
	CLK_TOP_ULPOSC1_D4,
	CLK_TOP_ULPOSC1_D16
};

static const int usb_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D5_D4,
	CLK_TOP_UNIVPLL_D6_D4,
	CLK_TOP_UNIVPLL_D5_D2
};

static const int ssusb_xhci_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D5_D4,
	CLK_TOP_UNIVPLL_D6_D4,
	CLK_TOP_UNIVPLL_D5_D2
};

static const int usb_2p_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D5_D4,
	CLK_TOP_UNIVPLL_D6_D4,
	CLK_TOP_UNIVPLL_D5_D2
};

static const int ssusb_xhci_2p_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D5_D4,
	CLK_TOP_UNIVPLL_D6_D4,
	CLK_TOP_UNIVPLL_D5_D2
};

static const int usb_3p_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D5_D4,
	CLK_TOP_UNIVPLL_D6_D4,
	CLK_TOP_UNIVPLL_D5_D2
};

static const int ssusb_xhci_3p_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D5_D4,
	CLK_TOP_UNIVPLL_D6_D4,
	CLK_TOP_UNIVPLL_D5_D2
};

static const int i2c_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MAINPLL_D4_D8,
	CLK_TOP_UNIVPLL_D5_D4
};

static const int seninf_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D4_D4,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_UNIVPLL_D7,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_MMPLL_D6,
	CLK_TOP_UNIVPLL_D5
};

static const int seninf1_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D4_D4,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_UNIVPLL_D7,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_MMPLL_D6,
	CLK_TOP_UNIVPLL_D5
};

static const int gcpu_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MAINPLL_D6,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_MMPLL_D5_D2,
	CLK_TOP_UNIVPLL_D5_D2
};

static const int venc_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MMPLL_D4_D2,
	CLK_TOP_MAINPLL_D6,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_MMPLL_D6,
	CLK_TOP_MAINPLL_D5_D2,
	CLK_TOP_MAINPLL_D6_D2,
	CLK_TOP_MMPLL_D9,
	CLK_TOP_UNIVPLL_D4_D4,
	CLK_TOP_MAINPLL_D4,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_UNIVPLL_D5,
	CLK_TOP_UNIVPLL_D5_D2,
	CLK_TOP_MAINPLL_D5
};

static const int vdec_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MAINPLL_D5_D2,
	CLK_TOP_MMPLL_D6_D2,
	CLK_TOP_UNIVPLL_D5_D2,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_MMPLL_D4_D2,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_MAINPLL_D5,
	CLK_TOP_UNIVPLL_D5,
	CLK_TOP_MMPLL_D6,
	CLK_TOP_MAINPLL_D4,
	CLK_TOP_TVDPLL2,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_IMGPLL,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_MMPLL_D9
};

static const int pwm_parents[] = {
	CLK_TOP_CLK32K,
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D4_D8,
	CLK_TOP_UNIVPLL_D6_D4
};

static const int mcupm_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MAINPLL_D6_D2,
	CLK_TOP_MAINPLL_D7_D4
};

static const int spmi_p_mst_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_CLK13M,
	CLK_TOP_ULPOSC1_D8,
	CLK_TOP_ULPOSC1_D10,
	CLK_TOP_ULPOSC1_D16,
	CLK_TOP_ULPOSC1_D7,
	CLK_TOP_CLK32K,
	CLK_TOP_MAINPLL_D7_D8,
	CLK_TOP_MAINPLL_D6_D8,
	CLK_TOP_MAINPLL_D5_D8
};

static const int spmi_m_mst_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_CLK13M,
	CLK_TOP_ULPOSC1_D8,
	CLK_TOP_ULPOSC1_D10,
	CLK_TOP_ULPOSC1_D16,
	CLK_TOP_ULPOSC1_D7,
	CLK_TOP_CLK32K,
	CLK_TOP_MAINPLL_D7_D8,
	CLK_TOP_MAINPLL_D6_D8,
	CLK_TOP_MAINPLL_D5_D8
};

static const int dvfsrc_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_ULPOSC1_D10,
	CLK_TOP_UNIVPLL_D6_D8,
	CLK_TOP_MSDCPLL_D16
};

static const int tl_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D5_D4,
	CLK_TOP_MAINPLL_D4_D4
};

static const int aes_msdcfde_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_MAINPLL_D6,
	CLK_TOP_MAINPLL_D4_D4,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_UNIVPLL_D6
};

static const int dsi_occ_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_UNIVPLL_D5_D2,
	CLK_TOP_UNIVPLL_D4_D2
};

static const int wpe_vpp_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MAINPLL_D5_D2,
	CLK_TOP_MMPLL_D6_D2,
	CLK_TOP_UNIVPLL_D5_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_UNIVPLL_D4_D2,
	CLK_TOP_MMPLL_D4_D2,
	CLK_TOP_MAINPLL_D6,
	CLK_TOP_MMPLL_D7,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_MAINPLL_D5,
	CLK_TOP_UNIVPLL_D5,
	CLK_TOP_MAINPLL_D4,
	CLK_TOP_TVDPLL1,
	CLK_TOP_UNIVPLL_D4
};

static const int hdcp_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D4_D8,
	CLK_TOP_MAINPLL_D5_D8,
	CLK_TOP_UNIVPLL_D6_D4
};

static const int hdcp_24m_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_192M_D4,
	CLK_TOP_UNIVPLL_192M_D8,
	CLK_TOP_UNIVPLL_D6_D8
};

static const int hdmi_apb_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D4,
	CLK_TOP_MSDCPLL_D2
};

static const int snps_eth_250m_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_ETHPLL_D2
};

static const int snps_eth_62p4m_ptp_parents[] = {
	CLK_TOP_APLL2_D3,
	CLK_TOP_APLL1_D3,
	CLK_TOP_CLK26M,
	CLK_TOP_ETHPLL_D8
};

static const int snps_eth_50m_rmii_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_ETHPLL_D10
};

static const int adsp_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_CLK13M,
	CLK_TOP_MAINPLL_D6,
	CLK_TOP_MAINPLL_D5_D2,
	CLK_TOP_UNIVPLL_D4_D4,
	CLK_TOP_UNIVPLL_D4,
	CLK_TOP_ULPOSC1_D2,
	CLK_TOP_ULPOSC1_CK1,
	CLK_TOP_ADSPPLL,
	CLK_TOP_ADSPPLL_D2,
	CLK_TOP_ADSPPLL_D4,
	CLK_TOP_ADSPPLL_D8
};

static const int audio_local_bus_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_CLK13M,
	CLK_TOP_MAINPLL_D4_D4,
	CLK_TOP_MAINPLL_D7_D2,
	CLK_TOP_MAINPLL_D5_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_MAINPLL_D7,
	CLK_TOP_MAINPLL_D4,
	CLK_TOP_UNIVPLL_D6,
	CLK_TOP_ULPOSC1_CK1,
	CLK_TOP_ULPOSC1_D4,
	CLK_TOP_ULPOSC1_D2
};

static const int asm_h_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D4,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_MAINPLL_D5_D2
};

static const int asm_l_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_UNIVPLL_D6_D4,
	CLK_TOP_UNIVPLL_D6_D2,
	CLK_TOP_MAINPLL_D5_D2
};

static const int apll1_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_APLL1_D4
};

static const int apll2_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_APLL2_D4
};

static const int apll3_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_APLL3_D4
};

static const int apll4_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_APLL4_D4
};

static const int apll5_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_APLL5_D4
};

static const int i2so1_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_APLL1,
	CLK_TOP_APLL2,
	CLK_TOP_APLL3,
	CLK_TOP_APLL4,
	CLK_TOP_APLL5
};

static const int i2so2_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_APLL1,
	CLK_TOP_APLL2,
	CLK_TOP_APLL3,
	CLK_TOP_APLL4,
	CLK_TOP_APLL5
};

static const int i2si1_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_APLL1,
	CLK_TOP_APLL2,
	CLK_TOP_APLL3,
	CLK_TOP_APLL4,
	CLK_TOP_APLL5
};

static const int i2si2_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_APLL1,
	CLK_TOP_APLL2,
	CLK_TOP_APLL3,
	CLK_TOP_APLL4,
	CLK_TOP_APLL5
};

static const int dptx_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_APLL1,
	CLK_TOP_APLL2,
	CLK_TOP_APLL3,
	CLK_TOP_APLL4,
	CLK_TOP_APLL5
};

static const int aud_iec_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_APLL1,
	CLK_TOP_APLL2,
	CLK_TOP_APLL3,
	CLK_TOP_APLL4,
	CLK_TOP_APLL5
};

static const int a1sys_hp_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_APLL1_D4
};

static const int a2sys_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_APLL2_D4
};

static const int a3sys_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_APLL3_D4,
	CLK_TOP_APLL4_D4,
	CLK_TOP_APLL5_D4
};

static const int a4sys_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_APLL3_D4,
	CLK_TOP_APLL4_D4,
	CLK_TOP_APLL5_D4
};

static const int ecc_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_MAINPLL_D4_D4,
	CLK_TOP_MAINPLL_D5_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_MAINPLL_D6,
	CLK_TOP_UNIVPLL_D6
};

static const int spinor_parents[] = {
	CLK_TOP_CLK26M,
	CLK_TOP_CLK13M,
	CLK_TOP_MAINPLL_D7_D8,
	CLK_TOP_UNIVPLL_D6_D8
};

static const int ulposc_parents[] = {
	CLK_TOP_ULPOSC_CK1,
	CLK_TOP_ETHPLL_D2,
	CLK_TOP_MAINPLL_D4_D2,
	CLK_TOP_ETHPLL_D10
};

static const int srck_parents[] = {
	CLK_TOP_ULPOSC1_D10,
	CLK_TOP_CLK26M
};

static const struct mtk_composite top_muxes[] = {
	/* CLK_CFG_0 */
	MUX_GATE(CLK_TOP_AXI, axi_parents, 0x020, 0, 4, 7),
	MUX_GATE(CLK_TOP_SPM, spm_parents, 0x020, 8, 4, 15),
	MUX_GATE(CLK_TOP_SCP, scp_parents, 0x020, 16, 4, 23),
	MUX_GATE(CLK_TOP_BUS_AXIMEM, bus_aximem_parents, 0x020, 24, 4, 31),
	/* CLK_CFG_1 */
	MUX_GATE(CLK_TOP_VPP, vpp_parents, 0x02C, 0, 4, 7),
	MUX_GATE(CLK_TOP_ETHDR, ethdr_parents, 0x02C, 8, 4, 15),
	MUX_GATE(CLK_TOP_IPE, ipe_parents, 0x02C, 16, 4, 23),
	MUX_GATE(CLK_TOP_CAM, cam_parents, 0x02C, 24, 4, 31),
	/* CLK_CFG_2 */
	MUX_GATE(CLK_TOP_CCU, ccu_parents, 0x038, 0, 4, 7),
	MUX_GATE(CLK_TOP_CCU_AHB, ccu_ahb_parents, 0x038, 8, 4, 15),
	MUX_GATE(CLK_TOP_IMG, img_parents, 0x038, 16, 4, 23),
	MUX_GATE(CLK_TOP_CAMTM, camtm_parents, 0x038, 24, 4, 31),
	/* CLK_CFG_3 */
	MUX_GATE(CLK_TOP_DSP, dsp_parents, 0x044, 0, 4, 7),
	MUX_GATE(CLK_TOP_DSP1, dsp1_parents, 0x044, 8, 4, 15),
	MUX_GATE(CLK_TOP_DSP2, dsp2_parents, 0x044, 16, 4, 23),
	MUX_GATE(CLK_TOP_DSP3, dsp3_parents, 0x044, 24, 4, 31),
	/* CLK_CFG_4 */
	MUX_GATE(CLK_TOP_DSP4, dsp4_parents, 0x050, 0, 4, 7),
	MUX_GATE(CLK_TOP_DSP5, dsp5_parents, 0x050, 8, 4, 15),
	MUX_GATE(CLK_TOP_DSP6, dsp6_parents, 0x050, 16, 4, 23),
	MUX_GATE(CLK_TOP_DSP7, dsp7_parents, 0x050, 24, 4, 31),
	/* CLK_CFG_5 */
	MUX_GATE(CLK_TOP_MFG_CORE_TMP, mfg_core_tmp_parents, 0x05C, 0, 4, 7),
	MUX_GATE(CLK_TOP_CAMTG, camtg_parents, 0x05C, 8, 4, 15),
	MUX_GATE(CLK_TOP_CAMTG2, camtg2_parents, 0x05C, 16, 4, 23),
	MUX_GATE(CLK_TOP_CAMTG3, camtg3_parents, 0x05C, 24, 4, 31),
	/* CLK_CFG_6 */
	MUX_GATE(CLK_TOP_UART, uart_parents, 0x068, 0, 4, 7),
	MUX_GATE(CLK_TOP_SPI, spi_parents, 0x068, 8, 4, 15),
	MUX_GATE(CLK_TOP_MSDC50_0_HCLK, msdc5hclk_parents, 0x068, 16, 4, 23),
	MUX_GATE(CLK_TOP_MSDC50_0, msdc50_0_parents, 0x068, 24, 4, 31),
	/* CLK_CFG_7 */
	MUX_GATE(CLK_TOP_MSDC30_1, msdc30_1_parents, 0x074, 0, 4, 7),
	MUX_GATE(CLK_TOP_MSDC30_2, msdc30_2_parents, 0x074, 8, 4, 15),
	MUX_GATE(CLK_TOP_INTDIR, intdir_parents, 0x074, 16, 4, 23),
	MUX_GATE(CLK_TOP_AUD_INTBUS, aud_intbus_parents, 0x074, 24, 4, 31),
	/* CLK_CFG_8 */
	MUX_GATE(CLK_TOP_AUDIO_H, audio_h_parents, 0x080, 0, 4, 7),
	MUX_GATE(CLK_TOP_PWRAP_ULPOSC, pwrap_ulposc_parents, 0x080, 8, 4, 15),
	MUX_GATE(CLK_TOP_ATB, atb_parents, 0x080, 16, 4, 23),
	MUX_GATE(CLK_TOP_SSPM, sspm_parents, 0x080, 24, 4, 31),
	/* CLK_CFG_9 */
	MUX_GATE(CLK_TOP_DP, dp_parents, 0x08C, 0, 4, 7),
	MUX_GATE(CLK_TOP_EDP, edp_parents, 0x08C, 8, 4, 15),
	MUX_GATE(CLK_TOP_DPI, dpi_parents, 0x08C, 16, 4, 23),
	MUX_GATE(CLK_TOP_DISP_PWM0, disp_pwm0_parents, 0x08C, 24, 4, 31),
	/* CLK_CFG_10 */
	MUX_GATE(CLK_TOP_DISP_PWM1, disp_pwm1_parents, 0x098, 0, 4, 7),
	MUX_GATE(CLK_TOP_USB_TOP, usb_parents, 0x098, 8, 4, 15),
	MUX_GATE(CLK_TOP_SSUSB_XHCI, ssusb_xhci_parents, 0x098, 16, 4, 23),
	MUX_GATE(CLK_TOP_USB_TOP_2P, usb_2p_parents, 0x098, 24, 4, 31),
	/* CLK_CFG_11 */
	MUX_GATE(CLK_TOP_SSUSB_XHCI_2P, ssusb_xhci_2p_parents, 0x0A4, 0, 4, 7),
	MUX_GATE(CLK_TOP_USB_TOP_3P, usb_3p_parents, 0x0A4, 8, 4, 15),
	MUX_GATE(CLK_TOP_SSUSB_XHCI_3P, ssusb_xhci_3p_parents, 0x0A4, 16, 4, 23),
	MUX_GATE(CLK_TOP_I2C, i2c_parents, 0x0A4, 24, 4, 31),
	/* CLK_CFG_12 */
	MUX_GATE(CLK_TOP_SENINF, seninf_parents, 0x0B0, 0, 4, 7),
	MUX_GATE(CLK_TOP_SENINF1, seninf1_parents, 0x0B0, 8, 4, 15),
	MUX_GATE(CLK_TOP_GCPU, gcpu_parents, 0x0B0, 16, 4, 23),
	MUX_GATE(CLK_TOP_VENC, venc_parents, 0x0B0, 24, 4, 31),
	/* CLK_CFG_13 */
	MUX_GATE(CLK_TOP_VDEC, vdec_parents, 0x0BC, 0, 4, 7),
	MUX_GATE(CLK_TOP_PWM, pwm_parents, 0x0BC, 8, 4, 15),
	MUX_GATE(CLK_TOP_MCUPM, mcupm_parents, 0x0BC, 16, 4, 23),
	MUX_GATE(CLK_TOP_SPMI_P_MST, spmi_p_mst_parents, 0x0BC, 24, 4, 31),
	/* CLK_CFG_14 */
	MUX_GATE(CLK_TOP_SPMI_M_MST, spmi_m_mst_parents, 0x0C8, 0, 4, 7),
	MUX_GATE(CLK_TOP_DVFSRC, dvfsrc_parents, 0x0C8, 8, 4, 15),
	MUX_GATE(CLK_TOP_TL, tl_parents, 0x0C8, 16, 4, 23),
	MUX_GATE(CLK_TOP_AES_MSDCFDE, aes_msdcfde_parents, 0x0C8, 24, 4, 31),
	/* CLK_CFG_15 */
	MUX_GATE(CLK_TOP_DSI_OCC, dsi_occ_parents, 0x0D4, 0, 4, 7),
	MUX_GATE(CLK_TOP_WPE_VPP, wpe_vpp_parents, 0x0D4, 8, 4, 15),
	MUX_GATE(CLK_TOP_HDCP, hdcp_parents, 0x0D4, 16, 4, 23),
	MUX_GATE(CLK_TOP_HDCP_24M, hdcp_24m_parents, 0x0D4, 24, 4, 31),
	/* CLK_CFG_16 */
	MUX_GATE(CLK_TOP_HDMI_APB, hdmi_apb_parents, 0x0E0, 0, 4, 7),
	MUX_GATE(CLK_TOP_SNPS_ETH_250M, snps_eth_250m_parents, 0x0E0, 8, 4, 15),
	MUX_GATE(CLK_TOP_SNPS_ETH_62P4M_PTP, snps_eth_62p4m_ptp_parents, 0x0E0, 16, 4, 23),
	MUX_GATE(CLK_TOP_SNPS_ETH_50M_RMII, snps_eth_50m_rmii_parents, 0x0E0, 24, 4, 31),
	/* CLK_CFG_17 */
	MUX_GATE(CLK_TOP_ADSP, adsp_parents, 0x0EC, 0, 4, 7),
	MUX_GATE(CLK_TOP_AUDIO_LOCAL_BUS, audio_local_bus_parents, 0x0EC, 8, 4, 15),
	MUX_GATE(CLK_TOP_ASM_H, asm_h_parents, 0x0EC, 16, 4, 23),
	MUX_GATE(CLK_TOP_ASM_L, asm_l_parents, 0x0EC, 24, 4, 31),
	/* CLK_CFG_18 */
	MUX_GATE(CLK_TOP_APLL1, apll1_parents, 0x0F8, 0, 4, 7),
	MUX_GATE(CLK_TOP_APLL2, apll2_parents, 0x0F8, 8, 4, 15),
	MUX_GATE(CLK_TOP_APLL3, apll3_parents, 0x0F8, 16, 4, 23),
	MUX_GATE(CLK_TOP_APLL4, apll4_parents, 0x0F8, 24, 4, 31),
	/* CLK_CFG_19 */
	MUX_GATE(CLK_TOP_APLL5, apll5_parents, 0x0104, 0, 4, 7),
	MUX_GATE(CLK_TOP_I2SO1, i2so1_parents, 0x0104, 8, 4, 15),
	MUX_GATE(CLK_TOP_I2SO2, i2so2_parents, 0x0104, 16, 4, 23),
	MUX_GATE(CLK_TOP_I2SI1, i2si1_parents, 0x0104, 24, 4, 31),
	/* CLK_CFG_20 */
	MUX_GATE(CLK_TOP_I2SI2, i2si2_parents, 0x0110, 0, 4, 7),
	MUX_GATE(CLK_TOP_DPTX, dptx_parents, 0x0110, 8, 4, 15),
	MUX_GATE(CLK_TOP_AUD_IEC, aud_iec_parents, 0x0110, 16, 4, 23),
	MUX_GATE(CLK_TOP_A1SYS_HP, a1sys_hp_parents, 0x0110, 24, 4, 31),
	/* CLK_CFG_21 */
	MUX_GATE(CLK_TOP_A2SYS, a2sys_parents, 0x011C, 0, 4, 7),
	MUX_GATE(CLK_TOP_A3SYS, a3sys_parents, 0x011C, 8, 4, 15),
	MUX_GATE(CLK_TOP_A4SYS, a4sys_parents, 0x011C, 16, 4, 23),
	MUX_GATE(CLK_TOP_ECC, ecc_parents, 0x011C, 24, 4, 31),
	/* CLK_CFG_22 */
	MUX_GATE(CLK_TOP_SPINOR, spinor_parents, 0x0128, 0, 4, 7),
	MUX_GATE(CLK_TOP_ULPOSC, ulposc_parents, 0x0128, 8, 4, 15),
	MUX_GATE(CLK_TOP_SRCK, srck_parents, 0x0128, 16, 4, 23),
};

static const struct mtk_composite * find_top_mux(unsigned long id)
{
	const struct mtk_composite *mux = NULL;
	u32 i;

	for (i = 0; i < ARRAY_SIZE(top_muxes); i++) {
		if (id == top_muxes[i].id) {
			mux = &top_muxes[i];
			break;
		}
	}

	return mux;
}

static const struct mtk_clk_tree mt8188_clk_tree = {
	.xtal_rate = 26 * MHZ,
	.xtal2_rate = 26 * MHZ,
	.plls = apmixed_plls,
	/* Several ops are re-implemented in this driver.
	 *  The following fields doesn't need to be set:
	 * .fdivs_offs
	 * .muxes_offs
	 * .fclks
	 * .fdivs
	 * .muxes
	 */
};

static const struct mtk_gate_regs infra_ao0_cg_regs = {
	.set_ofs = 0x80,
	.clr_ofs = 0x84,
	.sta_ofs = 0x90,
};

static const struct mtk_gate_regs infra_ao1_cg_regs = {
	.set_ofs = 0x88,
	.clr_ofs = 0x8c,
	.sta_ofs = 0x94,
};

static const struct mtk_gate_regs infra_ao2_cg_regs = {
	.set_ofs = 0xa4,
	.clr_ofs = 0xa8,
	.sta_ofs = 0xac,
};

static const struct mtk_gate_regs infra_ao3_cg_regs = {
	.set_ofs = 0xc0,
	.clr_ofs = 0xc4,
	.sta_ofs = 0xc8,
};

static const struct mtk_gate_regs infra_ao4_cg_regs = {
	.set_ofs = 0xe0,
	.clr_ofs = 0xe4,
	.sta_ofs = 0xe8,
};

#define GATE_INFRA_AO0(_id, _parent, _shift) {			\
		.id = _id,					\
		.parent = _parent,				\
		.regs = &infra_ao0_cg_regs,			\
		.shift = _shift,				\
		.flags = CLK_GATE_SETCLR | CLK_PARENT_TOPCKGEN,	\
	}

#define GATE_INFRA_AO1(_id, _parent, _shift) {			\
		.id = _id,					\
		.parent = _parent,				\
		.regs = &infra_ao1_cg_regs,			\
		.shift = _shift,				\
		.flags = CLK_GATE_SETCLR | CLK_PARENT_TOPCKGEN,	\
	}

#define GATE_INFRA_AO2(_id, _parent, _shift) {			\
		.id = _id,					\
		.parent = _parent,				\
		.regs = &infra_ao2_cg_regs,			\
		.shift = _shift,				\
		.flags = CLK_GATE_SETCLR | CLK_PARENT_TOPCKGEN,	\
	}

#define GATE_INFRA_AO3(_id, _parent, _shift) {			\
		.id = _id,					\
		.parent = _parent,				\
		.regs = &infra_ao3_cg_regs,			\
		.shift = _shift,				\
		.flags = CLK_GATE_SETCLR | CLK_PARENT_TOPCKGEN,	\
	}

#define GATE_INFRA_AO4(_id, _parent, _shift) {			\
		.id = _id,					\
		.parent = _parent,				\
		.regs = &infra_ao4_cg_regs,			\
		.shift = _shift,				\
		.flags = CLK_GATE_SETCLR | CLK_PARENT_TOPCKGEN,	\
	}

static const struct mtk_gate infracfg_ao_clks[] = {
	/* INFRA_AO0 */
	GATE_INFRA_AO0(CLK_INFRA_AO_PMIC_TMR, CLK_TOP_PWRAP_ULPOSC, 0),
	GATE_INFRA_AO0(CLK_INFRA_AO_PMIC_AP, CLK_TOP_PWRAP_ULPOSC, 1),
	GATE_INFRA_AO0(CLK_INFRA_AO_PMIC_MD, CLK_TOP_PWRAP_ULPOSC, 2),
	GATE_INFRA_AO0(CLK_INFRA_AO_PMIC_CONN, CLK_TOP_PWRAP_ULPOSC, 3),
	/* infra_ao_sej is main clock is for secure engine with JTAG support */
	GATE_INFRA_AO0(CLK_INFRA_AO_SEJ, CLK_TOP_AXI, 5),
	GATE_INFRA_AO0(CLK_INFRA_AO_APXGPT, CLK_TOP_AXI, 6),
	GATE_INFRA_AO0(CLK_INFRA_AO_GCE, CLK_TOP_AXI, 8),
	GATE_INFRA_AO0(CLK_INFRA_AO_GCE2, CLK_TOP_AXI, 9),
	GATE_INFRA_AO0(CLK_INFRA_AO_THERM, CLK_TOP_AXI, 10),
	GATE_INFRA_AO0(CLK_INFRA_AO_PWM_HCLK, CLK_TOP_AXI, 15),
	GATE_INFRA_AO0(CLK_INFRA_AO_PWM1, CLK_TOP_PWM, 16),
	GATE_INFRA_AO0(CLK_INFRA_AO_PWM2, CLK_TOP_PWM, 17),
	GATE_INFRA_AO0(CLK_INFRA_AO_PWM3, CLK_TOP_PWM, 18),
	GATE_INFRA_AO0(CLK_INFRA_AO_PWM4, CLK_TOP_PWM, 19),
	GATE_INFRA_AO0(CLK_INFRA_AO_PWM, CLK_TOP_PWM, 21),
	GATE_INFRA_AO0(CLK_INFRA_AO_UART0, CLK_TOP_UART, 22),
	GATE_INFRA_AO0(CLK_INFRA_AO_UART1, CLK_TOP_UART, 23),
	GATE_INFRA_AO0(CLK_INFRA_AO_UART2, CLK_TOP_UART, 24),
	GATE_INFRA_AO0(CLK_INFRA_AO_UART3, CLK_TOP_UART, 25),
	GATE_INFRA_AO0(CLK_INFRA_AO_UART4, CLK_TOP_UART, 26),
	GATE_INFRA_AO0(CLK_INFRA_AO_GCE_26M, CLK_TOP_CLK26M, 27),
	GATE_INFRA_AO0(CLK_INFRA_AO_UART5, CLK_TOP_UART, 29),
	/* INFRA_AO1 */
	GATE_INFRA_AO1(CLK_INFRA_AO_HDMI_26M, CLK_TOP_CLK26M, 0),
	GATE_INFRA_AO1(CLK_INFRA_AO_SPI0, CLK_TOP_SPI, 1),
	GATE_INFRA_AO1(CLK_INFRA_AO_MSDC0, CLK_TOP_CLK26M, 2),
	GATE_INFRA_AO1(CLK_INFRA_AO_MSDC1, CLK_TOP_AXI, 4),
	GATE_INFRA_AO1(CLK_INFRA_AO_MSDC2, CLK_TOP_AXI, 5),
	GATE_INFRA_AO1(CLK_INFRA_AO_MSDC0_SRC, CLK_TOP_MSDC50_0, 6),
	/* infra_ao_dvfsrc is for internal DVFS usage, should not be handled by Linux. */
	GATE_INFRA_AO1(CLK_INFRA_AO_DVFSRC, CLK_TOP_CLK26M, 7),
	GATE_INFRA_AO1(CLK_INFRA_AO_TRNG, CLK_TOP_AXI, 9),
	GATE_INFRA_AO1(CLK_INFRA_AO_AUXADC, CLK_TOP_CLK26M, 10),
	GATE_INFRA_AO1(CLK_INFRA_AO_CPUM, CLK_TOP_AXI, 11),
	GATE_INFRA_AO1(CLK_INFRA_AO_HDMI_32K, CLK_TOP_CLK32K, 12),
	GATE_INFRA_AO1(CLK_INFRA_AO_CEC_66M_HCLK, CLK_TOP_AXI, 13),
	GATE_INFRA_AO1(CLK_INFRA_AO_PCIE_TL_26M, CLK_TOP_CLK26M, 15),
	GATE_INFRA_AO1(CLK_INFRA_AO_MSDC1_SRC, CLK_TOP_MSDC30_1, 16),
	GATE_INFRA_AO1(CLK_INFRA_AO_CEC_66M_BCLK, CLK_TOP_AXI, 17),
	GATE_INFRA_AO1(CLK_INFRA_AO_PCIE_TL_96M, CLK_TOP_TL, 18),
	/* infra_ao_dapc is for device access permission control module */
	GATE_INFRA_AO1(CLK_INFRA_AO_DEVICE_APC, CLK_TOP_AXI, 20),
	GATE_INFRA_AO1(CLK_INFRA_AO_ECC_66M_HCLK, CLK_TOP_AXI, 23),
	GATE_INFRA_AO1(CLK_INFRA_AO_DEBUGSYS, CLK_TOP_AXI, 24),
	GATE_INFRA_AO1(CLK_INFRA_AO_AUDIO, CLK_TOP_AXI, 25),
	GATE_INFRA_AO1(CLK_INFRA_AO_PCIE_TL_32K, CLK_TOP_CLK32K, 26),
	GATE_INFRA_AO1(CLK_INFRA_AO_DBG_TRACE, CLK_TOP_AXI, 29),
	GATE_INFRA_AO1(CLK_INFRA_AO_DRAMC_F26M, CLK_TOP_CLK26M, 31),
	/* INFRA_AO2 */
	GATE_INFRA_AO2(CLK_INFRA_AO_IRTX, CLK_TOP_AXI, 0),
	GATE_INFRA_AO2(CLK_INFRA_AO_DISP_PWM, CLK_TOP_DISP_PWM0, 2),
	GATE_INFRA_AO2(CLK_INFRA_AO_CLDMA_BCLK, CLK_TOP_AXI, 3),
	GATE_INFRA_AO2(CLK_INFRA_AO_AUDIO_26M_BCLK, CLK_TOP_CLK26M, 4),
	GATE_INFRA_AO2(CLK_INFRA_AO_SPI1, CLK_TOP_SPI, 6),
	GATE_INFRA_AO2(CLK_INFRA_AO_SPI2, CLK_TOP_SPI, 9),
	GATE_INFRA_AO2(CLK_INFRA_AO_SPI3, CLK_TOP_SPI, 10),
	GATE_INFRA_AO2(CLK_INFRA_AO_FSSPM, CLK_TOP_SSPM, 15),
	GATE_INFRA_AO2(CLK_INFRA_AO_SSPM_BUS_HCLK, CLK_TOP_AXI, 17),
	GATE_INFRA_AO2(CLK_INFRA_AO_APDMA_BCLK, CLK_TOP_AXI, 18),
	GATE_INFRA_AO2(CLK_INFRA_AO_SPI4, CLK_TOP_SPI, 25),
	GATE_INFRA_AO2(CLK_INFRA_AO_SPI5, CLK_TOP_SPI, 26),
	GATE_INFRA_AO2(CLK_INFRA_AO_CQ_DMA, CLK_TOP_AXI, 27),
	/* INFRA_AO3 */
	GATE_INFRA_AO3(CLK_INFRA_AO_MSDC0_SELF, CLK_TOP_MSDC50_0, 0),
	GATE_INFRA_AO3(CLK_INFRA_AO_MSDC1_SELF, CLK_TOP_MSDC50_0, 1),
	GATE_INFRA_AO3(CLK_INFRA_AO_MSDC2_SELF, CLK_TOP_MSDC50_0, 2),
	GATE_INFRA_AO3(CLK_INFRA_AO_I2S_DMA, CLK_TOP_AXI, 5),
	GATE_INFRA_AO3(CLK_INFRA_AO_AP_MSDC0, CLK_TOP_MSDC50_0, 7),
	GATE_INFRA_AO3(CLK_INFRA_AO_MD_MSDC0, CLK_TOP_MSDC50_0, 8),
	GATE_INFRA_AO3(CLK_INFRA_AO_MSDC30_2, CLK_TOP_MSDC30_2, 9),
	GATE_INFRA_AO3(CLK_INFRA_AO_GCPU, CLK_TOP_GCPU, 10),
	GATE_INFRA_AO3(CLK_INFRA_AO_PCIE_PERI_26M, CLK_TOP_CLK26M, 15),
	GATE_INFRA_AO3(CLK_INFRA_AO_GCPU_66M_BCLK, CLK_TOP_AXI, 16),
	GATE_INFRA_AO3(CLK_INFRA_AO_GCPU_133M_BCLK, CLK_TOP_AXI, 17),
	GATE_INFRA_AO3(CLK_INFRA_AO_DISP_PWM1, CLK_TOP_DISP_PWM1, 20),
	GATE_INFRA_AO3(CLK_INFRA_AO_FBIST2FPC, CLK_TOP_MSDC50_0, 24),
	/* infra_ao_dapc_sync is for device access permission control module */
	GATE_INFRA_AO3(CLK_INFRA_AO_DEVICE_APC_SYNC, CLK_TOP_AXI, 25),
	GATE_INFRA_AO3(CLK_INFRA_AO_PCIE_P1_PERI_26M, CLK_TOP_CLK26M, 26),
	/* INFRA_AO4 */
	/* infra_ao_133m_mclk_set/infra_ao_66m_mclk_set are main clocks of peripheral */
	GATE_INFRA_AO4(CLK_INFRA_AO_133M_MCLK_CK, CLK_TOP_AXI, 0),
	GATE_INFRA_AO4(CLK_INFRA_AO_66M_MCLK_CK, CLK_TOP_AXI, 1),
	GATE_INFRA_AO4(CLK_INFRA_AO_RG_AES_MSDCFDE_CK_0P, CLK_TOP_AES_MSDCFDE, 18),
};

static const struct mtk_gate * find_infracfg_ao_gate(unsigned long id)
{
	const struct mtk_gate *gate = NULL;
	u32 i;

	for (i = 0; i < ARRAY_SIZE(infracfg_ao_clks); i++) {
		if (id == infracfg_ao_clks[i].id) {
			gate = &infracfg_ao_clks[i];
			break;
		}
	}

	return gate;
}

static const struct mtk_gate_regs peri_ao_cg_regs = {
	.set_ofs = 0x10,
	.clr_ofs = 0x14,
	.sta_ofs = 0x18,
};

#define GATE_PERI_AO(_id, _parent, _shift) {			\
		.id = _id,					\
		.parent = _parent,				\
		.regs = &peri_ao_cg_regs,			\
		.shift = _shift,				\
		.flags = CLK_GATE_SETCLR | CLK_PARENT_TOPCKGEN,	\
	}

static const struct mtk_gate pericfg_ao_clks[] = {
	GATE_PERI_AO(CLK_PERI_AO_ETHERNET, CLK_TOP_AXI, 0),
	GATE_PERI_AO(CLK_PERI_AO_ETHERNET_BUS, CLK_TOP_AXI, 1),
	GATE_PERI_AO(CLK_PERI_AO_FLASHIF_BUS, CLK_TOP_AXI, 3),
	GATE_PERI_AO(CLK_PERI_AO_FLASHIF_26M, CLK_TOP_CLK26M, 4),
	GATE_PERI_AO(CLK_PERI_AO_FLASHIFLASHCK, CLK_TOP_SPINOR, 5),
	GATE_PERI_AO(CLK_PERI_AO_SSUSB_2P_BUS, CLK_TOP_USB_TOP_2P, 9),
	GATE_PERI_AO(CLK_PERI_AO_SSUSB_2P_XHCI, CLK_TOP_SSUSB_XHCI_2P, 10),
	GATE_PERI_AO(CLK_PERI_AO_SSUSB_3P_BUS, CLK_TOP_USB_TOP_3P, 11),
	GATE_PERI_AO(CLK_PERI_AO_SSUSB_3P_XHCI, CLK_TOP_SSUSB_XHCI_3P, 12),
	GATE_PERI_AO(CLK_PERI_AO_SSUSB_BUS, CLK_TOP_USB_TOP, 13),
	GATE_PERI_AO(CLK_PERI_AO_SSUSB_XHCI, CLK_TOP_SSUSB_XHCI, 14),
	GATE_PERI_AO(CLK_PERI_AO_ETHERNET_MAC, CLK_TOP_SNPS_ETH_250M, 16),
};

static const struct mtk_gate * find_pericfg_ao_gate(unsigned long id)
{
	const struct mtk_gate *gate = NULL;
	u32 i;

	for (i = 0; i < ARRAY_SIZE(pericfg_ao_clks); i++) {
		if (id == pericfg_ao_clks[i].id) {
			gate = &pericfg_ao_clks[i];
			break;
		}
	}

	return gate;
}

static const struct mtk_gate_regs top0_cg_regs = {
	.set_ofs = 0x238,
	.clr_ofs = 0x238,
	.sta_ofs = 0x238,
};

static const struct mtk_gate_regs top1_cg_regs = {
	.set_ofs = 0x250,
	.clr_ofs = 0x250,
	.sta_ofs = 0x250,
};

#define GATE_TOP0(_id, _parent, _shift) {			   \
		.id = _id,					   \
		.parent = _parent,				   \
		.regs = &top0_cg_regs,				   \
		.shift = _shift,				   \
		.flags = CLK_GATE_NO_SETCLR | CLK_PARENT_TOPCKGEN, \
	}

#define GATE_TOP1(_id, _parent, _shift) {			       \
		.id = _id,					       \
		.parent = _parent,				       \
		.regs = &top1_cg_regs,				       \
		.shift = _shift,				       \
		.flags = CLK_GATE_NO_SETCLR_INV | CLK_PARENT_TOPCKGEN, \
	}

static const struct mtk_gate topckgen_cg_clks[] = {
	/* TOP0 */
	GATE_TOP0(CLK_TOP_CFGREG_CLOCK_EN_VPP0, CLK_TOP_VPP, 0),
	GATE_TOP0(CLK_TOP_CFGREG_CLOCK_EN_VPP1, CLK_TOP_VPP, 1),
	GATE_TOP0(CLK_TOP_CFGREG_CLOCK_EN_VDO0, CLK_TOP_VPP, 2),
	GATE_TOP0(CLK_TOP_CFGREG_CLOCK_EN_VDO1, CLK_TOP_VPP, 3),
	GATE_TOP0(CLK_TOP_CFGREG_CLOCK_ISP_AXI_GALS, CLK_TOP_VPP, 4),
	GATE_TOP0(CLK_TOP_CFGREG_F26M_VPP0, CLK_TOP_CLK26M, 5),
	GATE_TOP0(CLK_TOP_CFGREG_F26M_VPP1, CLK_TOP_CLK26M, 6),
	GATE_TOP0(CLK_TOP_CFGREG_F26M_VDO0, CLK_TOP_CLK26M, 7),
	GATE_TOP0(CLK_TOP_CFGREG_F26M_VDO1, CLK_TOP_CLK26M, 8),
	GATE_TOP0(CLK_TOP_CFGREG_AUD_F26M_AUD, CLK_TOP_CLK26M, 9),
	GATE_TOP0(CLK_TOP_CFGREG_UNIPLL_SES, CLK_TOP_UNIVPLL_D2, 15),
	GATE_TOP0(CLK_TOP_CFGREG_F_PCIE_PHY_REF, CLK_TOP_CLK26M, 18),
	/* TOP1 */
	GATE_TOP1(CLK_TOP_SSUSB_TOP_REF, CLK_TOP_CLK26M, 0),
	GATE_TOP1(CLK_TOP_SSUSB_PHY_REF, CLK_TOP_CLK26M, 1),
	GATE_TOP1(CLK_TOP_SSUSB_TOP_P1_REF, CLK_TOP_CLK26M, 2),
	GATE_TOP1(CLK_TOP_SSUSB_PHY_P1_REF, CLK_TOP_CLK26M, 3),
	GATE_TOP1(CLK_TOP_SSUSB_TOP_P2_REF, CLK_TOP_CLK26M, 4),
	GATE_TOP1(CLK_TOP_SSUSB_PHY_P2_REF, CLK_TOP_CLK26M, 5),
	GATE_TOP1(CLK_TOP_SSUSB_TOP_P3_REF, CLK_TOP_CLK26M, 6),
	GATE_TOP1(CLK_TOP_SSUSB_PHY_P3_REF, CLK_TOP_CLK26M, 7),
};

static const struct mtk_gate * find_topckgen_cg_gate(unsigned long id)
{
	const struct mtk_gate *gate = NULL;
	u32 i;

	for (i = 0; i < ARRAY_SIZE(topckgen_cg_clks); i++) {
		if (id == topckgen_cg_clks[i].id) {
			gate = &topckgen_cg_clks[i];
			break;
		}
	}

	return gate;
}

static int mt8188_pericfg_ao_enable(struct clk *clk)
{
	const struct mtk_gate *gate = find_pericfg_ao_gate(clk->id);

	if (!gate) {
		dev_err(clk->dev, "%s: failed to get pericfg gate: %lu\n", __func__, clk->id);
		return -ENODEV;
	}

	return mtk_common_clk_gate_enable(clk, gate);
}

static int mt8188_pericfg_ao_disable(struct clk *clk)
{
	const struct mtk_gate *gate = find_pericfg_ao_gate(clk->id);

	if (!gate) {
		dev_err(clk->dev, "%s: failed to get pericfg gate: %lu\n", __func__, clk->id);
		return -ENODEV;
	}

	return mtk_common_clk_gate_disable(clk, gate);
}

static ulong mt8188_pericfg_ao_get_rate(struct clk *clk)
{
	const struct mtk_gate *gate = find_pericfg_ao_gate(clk->id);

	if (!gate) {
		dev_err(clk->dev, "%s: failed to get pericfg gate: %lu\n", __func__, clk->id);
		return 0;
	}

	return mtk_common_clk_gate_get_rate(clk, gate);
}

static int mt8188_pericfg_ao_probe(struct udevice *dev)
{
	return mtk_common_clk_gate_init(dev, &mt8188_clk_tree, pericfg_ao_clks);
}

const struct clk_ops mt8188_pericfg_ao_ops = {
	.enable = mt8188_pericfg_ao_enable,
	.disable = mt8188_pericfg_ao_disable,
	.get_rate = mt8188_pericfg_ao_get_rate,
};

static int mt8188_infracfg_ao_enable(struct clk *clk)
{
	const struct mtk_gate *gate = find_infracfg_ao_gate(clk->id);

	if (!gate) {
		dev_err(clk->dev, "%s: failed to get pericfg gate: %lu\n", __func__, clk->id);
		return -ENODEV;
	}

	return mtk_common_clk_gate_enable(clk, gate);
}

static int mt8188_infracfg_ao_disable(struct clk *clk)
{
	const struct mtk_gate *gate = find_infracfg_ao_gate(clk->id);

	if (!gate) {
		dev_err(clk->dev, "%s: failed to get pericfg gate: %lu\n", __func__, clk->id);
		return -ENODEV;
	}

	return mtk_common_clk_gate_disable(clk, gate);
}

static ulong mt8188_infracfg_ao_get_rate(struct clk *clk)
{
	const struct mtk_gate *gate = find_infracfg_ao_gate(clk->id);

	if (!gate) {
		dev_err(clk->dev, "%s: failed to get pericfg gate: %lu\n", __func__, clk->id);
		return 0;
	}

	return mtk_common_clk_gate_get_rate(clk, gate);
}

static int mt8188_infracfg_ao_probe(struct udevice *dev)
{
	return mtk_common_clk_gate_init(dev, &mt8188_clk_tree, infracfg_ao_clks);
}

const struct clk_ops mt8188_infracfg_ao_ops = {
	.enable = mt8188_infracfg_ao_enable,
	.disable = mt8188_infracfg_ao_disable,
	.get_rate = mt8188_infracfg_ao_get_rate,
};

static int mt8188_topckgen_cg_enable(struct clk *clk)
{
	const struct mtk_gate *gate = find_topckgen_cg_gate(clk->id);

	if (!gate) {
		dev_err(clk->dev, "%s: failed to get pericfg gate: %lu\n", __func__, clk->id);
		return -ENODEV;
	}

	return mtk_common_clk_gate_enable(clk, gate);
}

static int mt8188_topckgen_cg_disable(struct clk *clk)
{
	const struct mtk_gate *gate = find_topckgen_cg_gate(clk->id);

	if (!gate) {
		dev_err(clk->dev, "%s: failed to get pericfg gate: %lu\n", __func__, clk->id);
		return -ENODEV;
	}

	return mtk_common_clk_gate_disable(clk, gate);
}

static ulong mt8188_topckgen_cg_get_rate(struct clk *clk)
{
	const struct mtk_gate *gate = find_topckgen_cg_gate(clk->id);

	if (!gate) {
		dev_err(clk->dev, "%s: failed to get pericfg gate: %lu\n", __func__, clk->id);
		return 0;
	}

	return mtk_common_clk_gate_get_rate(clk, gate);
}

static int mt8188_topckgen_cg_probe(struct udevice *dev)
{
	return mtk_common_clk_gate_init(dev, &mt8188_clk_tree, topckgen_cg_clks);
}

const struct clk_ops mt8188_topckgen_cg_ops = {
	.enable = mt8188_topckgen_cg_enable,
	.disable = mt8188_topckgen_cg_disable,
	.get_rate = mt8188_topckgen_cg_get_rate,
};

static int mt8188_topckgen_enable(struct clk *clk)
{
	struct mtk_clk_priv *priv = dev_get_priv(clk->dev);
	const struct mtk_composite *mux = find_top_mux(clk->id);
	u32 val;

	if (!mux) {
		dev_err(clk->dev, "%s: failed to get clk mux: %lu\n", __func__, clk->id);
		return -ENODEV;
	}

	/* enable clock gate */
	if (mux->flags & CLK_MUX_SETCLR_UPD) {
		val = BIT(mux->gate_shift);
		writel(val, priv->base + mux->mux_clr_reg);
	} else {
		val = readl(priv->base + mux->gate_reg);
		val &= ~BIT(mux->gate_shift);
		writel(val, priv->base + mux->gate_reg);
	}

	return 0;
}

static int mt8188_topckgen_disable(struct clk *clk)
{
	struct mtk_clk_priv *priv = dev_get_priv(clk->dev);
	const struct mtk_composite *mux = find_top_mux(clk->id);
	u32 val;

	if (!mux) {
		dev_err(clk->dev, "%s: failed to get clk mux: %lu\n", __func__, clk->id);
		return -ENODEV;
	}

	/* disable clock gate */
	if (mux->flags & CLK_MUX_SETCLR_UPD) {
		val = BIT(mux->gate_shift);
		writel(val, priv->base + mux->mux_set_reg);
	} else {
		val = readl(priv->base + mux->gate_reg);
		val |= BIT(mux->gate_shift);
		writel(val, priv->base + mux->gate_reg);
	}

	return 0;
}

static ulong mt8188_topckgen_get_rate(struct clk *clk)
{
	const struct mtk_fixed_clk *fclk = find_top_fixed_clk(clk->id);
	if (fclk)
		return fclk->rate;

	const struct mtk_fixed_factor *fdiv = find_top_fixed_div(clk->id);
	if (fdiv)
		return mtk_topckgen_get_factor_rate(clk, fdiv);

	const struct mtk_composite *mux = find_top_mux(clk->id);
	if (mux)
		return mtk_topckgen_get_mux_rate(clk, mux);

	dev_err(clk->dev, "%s: failed to get clk rate: %lu\n", __func__, clk->id);
	return -ENODEV;
}

static int mt8188_topckgen_set_parent(struct clk *clk, struct clk *parent)
{
	struct mtk_clk_priv *priv = dev_get_priv(clk->dev);

	const struct mtk_composite *mux = find_top_mux(clk->id);
	if (mux)
		return mtk_clk_mux_set_parent(priv->base, parent->id, mux);

	dev_err(clk->dev, "%s: failed to set clk parent: %lu\n", __func__, clk->id);
	return -ENODEV;
}

static int mt8188_topckgen_probe(struct udevice *dev)
{
	return mtk_common_clk_init(dev, &mt8188_clk_tree);
}

const struct clk_ops mt8188_topckgen_ops = {
	.enable = mt8188_topckgen_enable,
	.disable = mt8188_topckgen_disable,
	.get_rate = mt8188_topckgen_get_rate,
	.set_parent = mt8188_topckgen_set_parent,
};

static int mt8188_apmixedsys_probe(struct udevice *dev)
{
	return mtk_common_clk_init(dev, &mt8188_clk_tree);
}

static const struct udevice_id mt8188_apmixed_compat[] = {
	{ .compatible = "mediatek,mt8188-apmixedsys", },
	{ }
};

static const struct udevice_id mt8188_topckgen_compat[] = {
	{ .compatible = "mediatek,mt8188-topckgen", },
	{ }
};

static const struct udevice_id mt8188_topckgen_cg_compat[] = {
	{ .compatible = "mediatek,mt8188-topckgen-cg", },
	{ }
};

static const struct udevice_id mt8188_infracfg_ao_compat[] = {
	{ .compatible = "mediatek,mt8188-infracfg-ao", },
	{ }
};

static const struct udevice_id mt8188_pericfg_ao_compat[] = {
	{ .compatible = "mediatek,mt8188-pericfg-ao", },
	{ }
};

U_BOOT_DRIVER(mtk_clk_apmixedsys) = {
	.name = "mt8188-apmixedsys",
	.id = UCLASS_CLK,
	.of_match = mt8188_apmixed_compat,
	.probe = mt8188_apmixedsys_probe,
	.priv_auto = sizeof(struct mtk_clk_priv),
	.ops = &mtk_clk_apmixedsys_ops,
	.flags = DM_FLAG_PRE_RELOC,
};

U_BOOT_DRIVER(mtk_clk_topckgen) = {
	.name = "mt8188-topckgen",
	.id = UCLASS_CLK,
	.of_match = mt8188_topckgen_compat,
	.probe = mt8188_topckgen_probe,
	.priv_auto = sizeof(struct mtk_clk_priv),
	.ops = &mt8188_topckgen_ops,
	.flags = DM_FLAG_PRE_RELOC,
};

U_BOOT_DRIVER(mtk_clk_topckgen_cg) = {
	.name = "mt8188-topckgen-cg",
	.id = UCLASS_CLK,
	.of_match = mt8188_topckgen_cg_compat,
	.probe = mt8188_topckgen_cg_probe,
	.priv_auto = sizeof(struct mtk_cg_priv),
	.ops = &mt8188_topckgen_cg_ops,
	.flags = DM_FLAG_PRE_RELOC,
};

U_BOOT_DRIVER(mtk_clk_infracfg_ao) = {
	.name = "mt8188-infracfg-ao",
	.id = UCLASS_CLK,
	.of_match = mt8188_infracfg_ao_compat,
	.probe = mt8188_infracfg_ao_probe,
	.priv_auto = sizeof(struct mtk_cg_priv),
	.ops = &mt8188_infracfg_ao_ops,
	.flags = DM_FLAG_PRE_RELOC,
};

U_BOOT_DRIVER(mtk_clk_pericfg_ao) = {
	.name = "mt8188-pericfg-ao",
	.id = UCLASS_CLK,
	.of_match = mt8188_pericfg_ao_compat,
	.probe = mt8188_pericfg_ao_probe,
	.priv_auto = sizeof(struct mtk_cg_priv),
	.ops = &mt8188_pericfg_ao_ops,
	.flags = DM_FLAG_PRE_RELOC,
};
