// SPDX-License-Identifier: GPL-2.0
/*
 * Mediatek Display PWM
 *
 * Copyright (c) 2023 BayLibre, SAS.
 * Author: Julien Masson <jmasson@baylibre.com>
 */

#include <asm/io.h>
#include <clk.h>
#include <dm.h>
#include <dm/device_compat.h>
#include <linux/math64.h>
#include <pwm.h>

#define NSEC_PER_SEC 1000000000L

#define DISP_PWM_EN		0x00
#define DISP_PWM_CON_0		0x18
#define DISP_PWM_CON_1		0x1c
#define DISP_PWM_DEBUG		0x80

#define PWM_CLKDIV_SHIFT	16
#define PWM_CLKDIV_MAX		0x3ff
#define PWM_CLKDIV_MASK		(PWM_CLKDIV_MAX << PWM_CLKDIV_SHIFT)

#define PWM_PERIOD_BIT_WIDTH	12
#define PWM_PERIOD_MASK		((1 << PWM_PERIOD_BIT_WIDTH) - 1)

#define PWM_HIGH_WIDTH_SHIFT	16
#define PWM_HIGH_WIDTH_MASK	(0x1fff << PWM_HIGH_WIDTH_SHIFT)

struct mtk_disp_pwm_priv {
	void __iomem *base;
	struct udevice *dev;
	struct clk *main_clk;
	struct clk *mm_clk;
};

static void mtk_disp_pwm_update_bits(struct mtk_disp_pwm_priv *pwm, u32 offset,
				     u32 mask, u32 data)
{
	u32 tmp = readl(pwm->base + offset);

	writel((tmp & ~mask) | (data & mask), pwm->base + offset);
}

static int mtk_disp_pwm_set_config(struct udevice *dev, uint channel, uint period_ns,
				   uint duty_ns)
{
	struct mtk_disp_pwm_priv *pwm = dev_get_priv(dev);
	u32 clk_div, period, high_width, value;
	u64 div, rate;

	/*
	 * Find period, high_width and clk_div to suit duty_ns and period_ns.
	 * Calculate proper div value to keep period value in the bound.
	 *
	 * period_ns = 10^9 * (clk_div + 1) * (period + 1) / PWM_CLK_RATE
	 * duty_ns = 10^9 * (clk_div + 1) * high_width / PWM_CLK_RATE
	 *
	 * period = (PWM_CLK_RATE * period_ns) / (10^9 * (clk_div + 1)) - 1
	 * high_width = (PWM_CLK_RATE * duty_ns) / (10^9 * (clk_div + 1))
	 */
	rate = clk_get_rate(pwm->main_clk);
	clk_div = div_u64(rate * period_ns, NSEC_PER_SEC) >> PWM_PERIOD_BIT_WIDTH;
	if (clk_div > PWM_CLKDIV_MAX)
		return -EINVAL;

	div = NSEC_PER_SEC * (clk_div + 1);
	period = div64_u64(rate * period_ns, div);
	if (period > 0)
		period--;

	high_width = div64_u64(rate * duty_ns, div);
	value = period | (high_width << PWM_HIGH_WIDTH_SHIFT);

	mtk_disp_pwm_update_bits(pwm, DISP_PWM_CON_0, PWM_CLKDIV_MASK,
				 clk_div << PWM_CLKDIV_SHIFT);

	mtk_disp_pwm_update_bits(pwm, DISP_PWM_CON_1,
				 PWM_PERIOD_MASK | PWM_HIGH_WIDTH_MASK,
				 value);

	return 0;
}

static int mtk_disp_pwm_enable(struct udevice *dev)
{
	struct mtk_disp_pwm_priv *pwm = dev_get_priv(dev);
	int ret;

	ret = clk_enable(pwm->main_clk);
	if (ret) {
		dev_err(pwm->dev, "failed to enable main clk: %d\n", ret);
		return ret;
	}

	ret = clk_enable(pwm->mm_clk);
	if (ret) {
		dev_err(pwm->dev, "failed to enable mm clk: %d\n", ret);
		return ret;
	}

	mtk_disp_pwm_update_bits(pwm, DISP_PWM_EN, BIT(0), BIT(0));

	return ret;
}

static int mtk_disp_pwm_disable(struct udevice *dev)
{
	struct mtk_disp_pwm_priv *pwm = dev_get_priv(dev);

	mtk_disp_pwm_update_bits(pwm, DISP_PWM_EN, 0, 0);

	clk_disable(pwm->mm_clk);
	clk_disable(pwm->main_clk);

	return 0;
}

static int mtk_disp_pwm_set_enable(struct udevice *dev, uint channel, bool enable)
{
	return enable ? mtk_disp_pwm_enable(dev) : mtk_disp_pwm_disable(dev);
}

static int mtk_disp_pwm_probe(struct udevice *dev)
{
	struct mtk_disp_pwm_priv *pwm = dev_get_priv(dev);

	pwm->dev = dev;

	pwm->base = dev_remap_addr(dev);
	if (IS_ERR(pwm->base))
		return PTR_ERR(pwm->base);

	pwm->main_clk = devm_clk_get(dev, "main");
	if (IS_ERR(pwm->main_clk))
		return PTR_ERR(pwm->main_clk);

	pwm->mm_clk = devm_clk_get(dev, "mm");
	if (IS_ERR(pwm->mm_clk))
		return PTR_ERR(pwm->mm_clk);

	/* Disable interrupt request */
	mtk_disp_pwm_update_bits(pwm, DISP_PWM_CON_0, 0, 0);

	/* Enable CG for RX synchronizer circuit */
	mtk_disp_pwm_update_bits(pwm, DISP_PWM_DEBUG, BIT(0), BIT(0));

	/*
	 * Duty = Low level/period. When PWM controller is disabled,
	 * the output is 1.
	 */
	mtk_disp_pwm_update_bits(pwm, DISP_PWM_DEBUG, BIT(1), BIT(1));

	return 0;
}

static const struct pwm_ops mtk_disp_pwm_ops = {
	.set_config = mtk_disp_pwm_set_config,
	.set_enable = mtk_disp_pwm_set_enable,
};

static const struct udevice_id mtk_disp_pwm_ids[] = {
	{ .compatible = "mediatek,mt8365-disp-pwm" },
	{}
};

U_BOOT_DRIVER(mtk_disp_pwm) = {
	.name	   = "mtk_disp_pwm",
	.id	   = UCLASS_PWM,
	.of_match  = mtk_disp_pwm_ids,
	.ops	   = &mtk_disp_pwm_ops,
	.probe	   = mtk_disp_pwm_probe,
	.priv_auto = sizeof(struct mtk_disp_pwm_priv),
};
