// SPDX-License-Identifier: GPL-2.0
/*
 * Mediatek Video Adaptive Ambient Light support
 *
 * Copyright (c) 2023 BayLibre, SAS.
 * Author: Julien Masson <jmasson@baylibre.com>
 */

#include <asm/io.h>
#include <clk.h>
#include <dm.h>
#include <dm/device_compat.h>
#include <errno.h>

#define DISP_AAL_EN    0x00
#define AAL_EN		     BIT(0)

#define DISP_AAL_SIZE  0x30

struct mtk_disp_aal_priv {
	void __iomem *base;
	struct udevice *dev;
	struct clk clk;
};

void mtk_aal_config(struct udevice *dev, struct display_timing timing)
{
	struct mtk_disp_aal_priv *aal = dev_get_priv(dev);
	u32 width = timing.hactive.typ;
	u32 height = timing.vactive.typ;

	writel(height << 16 | width, aal->base + DISP_AAL_SIZE);
}

void mtk_aal_start(struct udevice *dev)
{
	struct mtk_disp_aal_priv *aal = dev_get_priv(dev);

	writel(AAL_EN, aal->base + DISP_AAL_EN);
}

int mtk_aal_enable(struct udevice *dev)
{
	struct mtk_disp_aal_priv *aal = dev_get_priv(dev);
	int ret;

	ret = clk_enable(&aal->clk);
	if (ret)
		dev_err(aal->dev, "failed to enable clk: %s\n", errno_str(ret));

	return ret;
}

int mtk_aal_disable(struct udevice *dev)
{
	struct mtk_disp_aal_priv *aal = dev_get_priv(dev);
	int ret;

	ret = clk_disable(&aal->clk);
	if (ret)
		dev_err(aal->dev, "failed to disable clk: %s\n", errno_str(ret));

	return ret;
}

static int mtk_disp_aal_probe(struct udevice *dev)
{
	struct mtk_disp_aal_priv *aal = dev_get_priv(dev);
	int ret;

	aal->dev = dev;

	aal->base = dev_remap_addr(dev);
	if (IS_ERR(aal->base))
		return PTR_ERR(aal->base);

	ret = clk_get_by_index(dev, 0, &aal->clk);
	if (ret)
		return ret;

	return 0;
}

static const struct udevice_id mtk_disp_aal_ids[] = {
	{ .compatible = "mediatek,mt8365-disp-aal" },
	{}
};

U_BOOT_DRIVER(mtk_disp_aal) = {
	.name	   = "mtk_disp_aal",
	.id	   = UCLASS_MISC,
	.of_match  = mtk_disp_aal_ids,
	.probe	   = mtk_disp_aal_probe,
	.priv_auto = sizeof(struct mtk_disp_aal_priv),
};
