// SPDX-License-Identifier: GPL-2.0
/*
 * Mediatek Video DSI support
 *
 * Copyright (c) 2023 BayLibre, SAS.
 * Author: Julien Masson <jmasson@baylibre.com>
 */

#include <asm/io.h>
#include <asm/system.h>
#include <backlight.h>
#include <clk.h>
#include <div64.h>
#include <dm.h>
#include <dm/device_compat.h>
#include <dsi_host.h>
#include <errno.h>
#include <linux/delay.h>
#include <linux/err.h>
#include <panel.h>
#include <video.h>
#include <video_bridge.h>

#include "mtk_ddp.h"
#include "mtk_disp_aal.h"
#include "mtk_disp_ccorr.h"
#include "mtk_disp_color.h"
#include "mtk_disp_dither.h"
#include "mtk_disp_gamma.h"
#include "mtk_disp_mutex.h"
#include "mtk_disp_ovl.h"
#include "mtk_disp_rdma.h"
#include "mtk_mipi_tx.h"

#define MTK_DSI_MAX_WIDTH  3840
#define MTK_DSI_MAX_HEIGHT 2160

#define DSI_START		0x00

#define DSI_CON_CTRL		0x10
#define DSI_EN			      BIT(1)

#define DSI_MODE_CTRL		0x14
#define CMD_MODE		      0
#define SYNC_PULSE_MODE		      1
#define SYNC_EVENT_MODE		      2
#define BURST_MODE		      3

#define DSI_TXRX_CTRL		0x18

#define DSI_PSCTRL		0x1c
#define DSI_PS_WC		      0x3fff
#define PACKED_PS_16BIT_RGB565	      (0 << 16)
#define LOOSELY_PS_18BIT_RGB666	      (1 << 16)
#define PACKED_PS_18BIT_RGB666	      (2 << 16)
#define PACKED_PS_24BIT_RGB888	      (3 << 16)

#define DSI_VSA_NL		0x20
#define DSI_VBP_NL		0x24
#define DSI_VFP_NL		0x28
#define DSI_VACT_NL		0x2C
#define DSI_SIZE_CON		0x38
#define DSI_HSA_WC		0x50
#define DSI_HBP_WC		0x54
#define DSI_HFP_WC		0x58
#define DSI_HSTX_CKL_WC		0x64

#define DSI_PHY_LCCON		0x104
#define LC_HS_TX_EN		       BIT(0)

#define DSI_SHADOW_DEBUG	0x190
#define FORCE_COMMIT		       BIT(0)
#define BYPASS_SHADOW		       BIT(1)

struct mtk_dsi_priv {
	void __iomem *base;
	struct udevice *dev;
	struct udevice *ovl;
	struct udevice *rdma;
	struct udevice *color;
	struct udevice *ccorr;
	struct udevice *aal;
	struct udevice *gamma;
	struct udevice *dither;
	struct udevice *mutex;
	struct udevice *mmsys;
	struct udevice *phy;
	struct udevice *panel;
	struct udevice *host;
	struct mipi_dsi_device mipi;
	struct clk *engine_clk;
	struct clk *digital_clk;
	struct clk *hs_clk;
	u64 data_rate;
};

static void mtk_dsi_mask(struct mtk_dsi_priv *dsi, u32 offset, u32 val, u32 mask)
{
	void __iomem *addr = dsi->base + offset;
	u32 tmp = readl(addr) & ~mask;

	tmp |= (val & mask);
	writel(tmp, addr);
}

static void mtk_dsi_shadow_debug(struct mtk_dsi_priv *dsi)
{
	writel(FORCE_COMMIT | BYPASS_SHADOW, dsi->base + DSI_SHADOW_DEBUG);
}

static void mtk_dsi_enable(struct mtk_dsi_priv *dsi)
{
	mtk_dsi_mask(dsi, DSI_CON_CTRL, DSI_EN, DSI_EN);
}

static bool mtk_dsi_clk_hs_state(struct mtk_dsi_priv *dsi)
{
	u32 tmp = readl(dsi->base + DSI_PHY_LCCON);

	return ((tmp & LC_HS_TX_EN) == 1) ? true : false;
}

static void mtk_dsi_clk_hs_mode(struct mtk_dsi_priv *dsi)
{
	if (!mtk_dsi_clk_hs_state(dsi))
		mtk_dsi_mask(dsi, DSI_PHY_LCCON, LC_HS_TX_EN, LC_HS_TX_EN);
}

static void mtk_dsi_rxtx_control(struct mtk_dsi_priv *dsi)
{
	struct mipi_dsi_panel_plat *plat = dev_get_plat(dsi->panel);
	u32 reg;

	switch (plat->lanes) {
	case 1:
		reg = 1 << 2;
		break;
	case 2:
		reg = 3 << 2;
		break;
	case 3:
		reg = 7 << 2;
		break;
	case 4:
		reg = 0xf << 2;
		break;
	default:
		reg = 0xf << 2;
		break;
	}

	reg |= (plat->mode_flags & MIPI_DSI_CLOCK_NON_CONTINUOUS) << 6;
	reg |= (plat->mode_flags & MIPI_DSI_MODE_EOT_PACKET) >> 3;

	writel(reg, dsi->base + DSI_TXRX_CTRL);
}

static void mtk_dsi_ps_control(struct mtk_dsi_priv *dsi,
			       struct display_timing timing)
{
	struct mipi_dsi_panel_plat *plat = dev_get_plat(dsi->panel);
	u32 bpp;
	u32 reg;

	switch (plat->format) {
	case MIPI_DSI_FMT_RGB888:
		reg = PACKED_PS_24BIT_RGB888;
		bpp = 3;
		break;
	case MIPI_DSI_FMT_RGB666:
		reg = LOOSELY_PS_18BIT_RGB666;
		bpp = 3;
		break;
	case MIPI_DSI_FMT_RGB666_PACKED:
		reg = PACKED_PS_18BIT_RGB666;
		bpp = 3;
		break;
	case MIPI_DSI_FMT_RGB565:
		reg = PACKED_PS_16BIT_RGB565;
		bpp = 2;
		break;
	default:
		reg = PACKED_PS_24BIT_RGB888;
		bpp = 3;
		break;
	}

	reg += timing.hactive.typ * bpp & DSI_PS_WC;
	writel(reg, dsi->base + DSI_PSCTRL);
}

static void mtk_dsi_config_vdo_timing(struct mtk_dsi_priv *dsi,
				      struct display_timing timing)
{
	struct mipi_dsi_panel_plat *plat = dev_get_plat(dsi->panel);
	u32 hsa, hbp, hfp, bpp;
	u32 data_phy_cycles = 3;

	bpp = (plat->format == MIPI_DSI_FMT_RGB565) ? 2 : 3;

	writel(timing.vsync_len.typ, dsi->base + DSI_VSA_NL);
	writel(timing.vback_porch.typ, dsi->base + DSI_VBP_NL);
	writel(timing.vfront_porch.typ, dsi->base + DSI_VFP_NL);
	writel(timing.vactive.typ, dsi->base + DSI_VACT_NL);

	writel(timing.vactive.typ << 16 | timing.hactive.typ, dsi->base + DSI_SIZE_CON);

	hsa = (timing.hsync_len.typ * bpp - 10);

	if (plat->mode_flags & MIPI_DSI_MODE_VIDEO_SYNC_PULSE)
		hbp = timing.hback_porch.typ * bpp - 10;
	else
		hbp = (timing.hback_porch.typ + timing.hsync_len.typ) * bpp - 10;

	if (plat->mode_flags & MIPI_DSI_MODE_VIDEO_BURST) {
		if ((timing.hfront_porch.typ + timing.hback_porch.typ) *
		    bpp > data_phy_cycles * plat->lanes + 18) {
			hfp = timing.hfront_porch.typ * bpp -
			      (data_phy_cycles * plat->lanes + 18) *
			      timing.hfront_porch.typ /
			      (timing.hfront_porch.typ + timing.hback_porch.typ);
			hbp = hbp - (data_phy_cycles * plat->lanes + 18) *
			      timing.hback_porch.typ /
			      (timing.hfront_porch.typ + timing.hback_porch.typ);
		} else {
			dev_warn(dsi->dev, "HFP less than d-phy, FPS will under 60Hz\n");
			hfp = timing.hfront_porch.typ * bpp;
		}
	} else {
		if ((timing.hfront_porch.typ + timing.hback_porch.typ) *
		    bpp > data_phy_cycles * plat->lanes + 12) {
			hfp = timing.hfront_porch.typ * bpp -
			      (data_phy_cycles * plat->lanes + 12) *
			      timing.hfront_porch.typ /
			      (timing.hfront_porch.typ + timing.hback_porch.typ);
			hbp = hbp - (data_phy_cycles * plat->lanes + 12) *
			      timing.hback_porch.typ /
			      (timing.hfront_porch.typ + timing.hback_porch.typ);
		} else {
			dev_warn(dsi->dev, "HFP less than d-phy, FPS will under 60Hz\n");
			hfp = timing.hfront_porch.typ * bpp;
		}
	}

	writel(hsa, dsi->base + DSI_HSA_WC);
	writel(hbp, dsi->base + DSI_HBP_WC);
	writel(hfp, dsi->base + DSI_HFP_WC);

	mtk_dsi_ps_control(dsi, timing);
}

static void mtk_dsi_set_display_mode(struct mtk_dsi_priv *dsi,
				     struct display_timing timing)
{
	struct video_priv *priv = dev_get_uclass_priv(dsi->dev);

	priv->xsize = timing.hactive.typ;
	priv->ysize = timing.vactive.typ;
	priv->bpix = VIDEO_BPP32;

	mtk_dsi_rxtx_control(dsi);
	udelay(100);
	mtk_dsi_config_vdo_timing(dsi, timing);
}

static void mtk_dsi_set_mode(struct mtk_dsi_priv *dsi)
{
	struct mipi_dsi_panel_plat *plat = dev_get_plat(dsi->panel);
	u32 vid_mode = CMD_MODE;

	if (plat->mode_flags & MIPI_DSI_MODE_VIDEO) {
		if (plat->mode_flags & MIPI_DSI_MODE_VIDEO_BURST)
			vid_mode = BURST_MODE;
		else if (plat->mode_flags & MIPI_DSI_MODE_VIDEO_SYNC_PULSE)
			vid_mode = SYNC_PULSE_MODE;
		else
			vid_mode = SYNC_EVENT_MODE;
	}

	writel(vid_mode, dsi->base + DSI_MODE_CTRL);
}

static void mtk_dsi_start(struct mtk_dsi_priv *dsi)
{
	writel(0, dsi->base + DSI_START);
	writel(1, dsi->base + DSI_START);
}

static void mtk_dsi_output_enable(struct mtk_dsi_priv *dsi)
{
	mtk_dsi_set_mode(dsi);

	mtk_dsi_clk_hs_mode(dsi);

	mtk_dsi_start(dsi);
}

static int mtk_dsi_power_on(struct mtk_dsi_priv *dsi, struct display_timing timing)
{
	struct mipi_dsi_panel_plat *plat = dev_get_plat(dsi->panel);
	int ret, bit_per_pixel;

	switch (plat->format) {
	case MIPI_DSI_FMT_RGB565:
		bit_per_pixel = 16;
		break;
	case MIPI_DSI_FMT_RGB666_PACKED:
		bit_per_pixel = 18;
		break;
	case MIPI_DSI_FMT_RGB666:
	case MIPI_DSI_FMT_RGB888:
	default:
		bit_per_pixel = 24;
		break;
	}

	dsi->data_rate = DIV_ROUND_UP_ULL(timing.pixelclock.typ * bit_per_pixel,
					  plat->lanes);

	mtk_mipi_tx_power_on(dsi->phy, dsi->data_rate);

	ret = clk_enable(dsi->engine_clk);
	if (ret) {
		dev_err(dsi->dev, "failed to enable engine clk: %d\n", ret);
		goto engine_clk_failed;
	}

	ret = clk_enable(dsi->digital_clk);
	if (ret) {
		dev_err(dsi->dev, "failed to enable digital clk: %d\n", ret);
		goto digital_clk_failed;
	}

	ret = mtk_ovl_enable(dsi->ovl);
	if (ret) {
		dev_err(dsi->dev, "failed to enable ovl: %d\n", ret);
		goto ovl_enable_failed;
	}

	ret = mtk_ovl_smi_enable(dsi->ovl);
	if (ret) {
		dev_err(dsi->dev, "failed to enable ovl smi: %d\n", ret);
		goto ovl_smi_enable_failed;
	}

	ret = mtk_rdma_enable(dsi->rdma);
	if (ret) {
		dev_err(dsi->dev, "failed to enable rdma: %d\n", ret);
		goto rdma_enable_failed;
	}

	ret = mtk_color_enable(dsi->color);
	if (ret) {
		dev_err(dsi->dev, "failed to enable color: %d\n", ret);
		goto color_enable_failed;
	}

	ret = mtk_ccorr_enable(dsi->ccorr);
	if (ret) {
		dev_err(dsi->dev, "failed to enable ccorr: %d\n", ret);
		goto ccorr_enable_failed;
	}

	ret = mtk_aal_enable(dsi->aal);
	if (ret) {
		dev_err(dsi->dev, "failed to enable aal: %d\n", ret);
		goto aal_enable_failed;
	}

	ret = mtk_gamma_enable(dsi->gamma);
	if (ret) {
		dev_err(dsi->dev, "failed to enable gamma: %d\n", ret);
		goto gamma_enable_failed;
	}

	ret = mtk_dither_enable(dsi->dither);
	if (ret) {
		dev_err(dsi->dev, "failed to enable dither: %d\n", ret);
		goto dither_enable_failed;
	}

	mtk_dsi_enable(dsi);

	mtk_dsi_shadow_debug(dsi);

	return 0;

dither_enable_failed:
	mtk_dither_disable(dsi->dither);
gamma_enable_failed:
	mtk_gamma_disable(dsi->gamma);
aal_enable_failed:
	mtk_aal_disable(dsi->aal);
ccorr_enable_failed:
	mtk_ccorr_disable(dsi->ccorr);
color_enable_failed:
	mtk_color_disable(dsi->color);
rdma_enable_failed:
	mtk_rdma_disable(dsi->rdma);
ovl_smi_enable_failed:
	mtk_ovl_smi_disable(dsi->ovl);
ovl_enable_failed:
	mtk_ovl_disable(dsi->ovl);
digital_clk_failed:
	clk_disable(dsi->digital_clk);
engine_clk_failed:
	clk_disable(dsi->engine_clk);

	mtk_mipi_tx_power_off(dsi->phy);
	return ret;
}

static int mtk_dsi_host_init(struct mtk_dsi_priv *dsi)
{
	struct mipi_dsi_panel_plat *plat = dev_get_plat(dsi->panel);

	plat->device = &dsi->mipi;

	return dsi_host_init(dsi->host, plat->device, NULL, 0, NULL);
}

static int mtk_dsi_probe(struct udevice *dev)
{
	struct mtk_dsi_priv *dsi = dev_get_priv(dev);
	struct video_uc_plat *plat = dev_get_uclass_plat(dev);
	struct display_timing timing;
	int ret;

	/* Before relocation we don't need to do anything */
	if (!(gd->flags & GD_FLG_RELOC))
		return 0;

	dsi->dev = dev;

	dsi->base = dev_remap_addr(dev);
	if (IS_ERR(dsi->base))
		return PTR_ERR(dsi->base);

	dsi->engine_clk = devm_clk_get(dev, "engine");
	if (IS_ERR(dsi->engine_clk))
		return PTR_ERR(dsi->engine_clk);

	dsi->digital_clk = devm_clk_get(dev, "digital");
	if (IS_ERR(dsi->digital_clk))
		return PTR_ERR(dsi->digital_clk);

	dsi->hs_clk = devm_clk_get(dev, "hs");
	if (IS_ERR(dsi->hs_clk))
		return PTR_ERR(dsi->hs_clk);

	ret = uclass_get_device_by_phandle(UCLASS_MISC, dev, "mediatek,ovl", &dsi->ovl);
	if (ret) {
		dev_err(dev, "cannot get ovl device: %s\n", errno_str(ret));
		return ret;
	}

	ret = uclass_get_device_by_phandle(UCLASS_MISC, dev, "mediatek,rdma", &dsi->rdma);
	if (ret) {
		dev_err(dev, "cannot get rdma device: %s\n", errno_str(ret));
		return ret;
	}

	ret = uclass_get_device_by_phandle(UCLASS_MISC, dev, "mediatek,color", &dsi->color);
	if (ret) {
		dev_err(dev, "cannot get color device: %s\n", errno_str(ret));
		return ret;
	}

	ret = uclass_get_device_by_phandle(UCLASS_MISC, dev, "mediatek,ccorr", &dsi->ccorr);
	if (ret) {
		dev_err(dev, "cannot get ccorr device: %s\n", errno_str(ret));
		return ret;
	}

	ret = uclass_get_device_by_phandle(UCLASS_MISC, dev, "mediatek,aal", &dsi->aal);
	if (ret) {
		dev_err(dev, "cannot get aal device: %s\n", errno_str(ret));
		return ret;
	}

	ret = uclass_get_device_by_phandle(UCLASS_MISC, dev, "mediatek,gamma", &dsi->gamma);
	if (ret) {
		dev_err(dev, "cannot get gamma device: %s\n", errno_str(ret));
		return ret;
	}

	ret = uclass_get_device_by_phandle(UCLASS_MISC, dev, "mediatek,dither", &dsi->dither);
	if (ret) {
		dev_err(dev, "cannot get dither device: %s\n", errno_str(ret));
		return ret;
	}

	ret = uclass_get_device_by_phandle(UCLASS_MISC, dev, "mediatek,mutex", &dsi->mutex);
	if (ret) {
		dev_err(dev, "cannot get mutex device: %s\n", errno_str(ret));
		return ret;
	}

	ret = uclass_get_device_by_phandle(UCLASS_MISC, dev, "mediatek,mmsys", &dsi->mmsys);
	if (ret) {
		dev_err(dev, "cannot get mmsys device: %s\n", errno_str(ret));
		return ret;
	}

	ret = uclass_get_device_by_phandle(UCLASS_MISC, dev, "phy", &dsi->phy);
	if (ret) {
		dev_err(dev, "cannot get phy device: %s\n", errno_str(ret));
		return ret;
	}

	ret = uclass_first_device_err(UCLASS_PANEL, &dsi->panel);
	if (ret) {
		dev_err(dev, "cannot get panel: %s\n", errno_str(ret));
		return ret;
	}

	ret = uclass_get_device(UCLASS_DSI_HOST, 0, &dsi->host);
	if (ret) {
		dev_err(dev, "cannot get dsi host: %s\n", errno_str(ret));
		return ret;
	}

	ret = mtk_dsi_host_init(dsi);
	if (ret) {
		dev_err(dev, "failed to initialize mipi dsi host\n");
		return ret;
	}

	mtk_disp_mutex_ovl_dsi_enable(dsi->mutex);

	mtk_ddp_ovl_to_dsi(dsi->mmsys);

	ret = panel_get_display_timing(dsi->panel, &timing);
	if (ret) {
		dev_err(dev, "cannot get display timing\n");
		return ret;
	}

	mtk_ovl_config(dsi->ovl, timing);
	mtk_ovl_start(dsi->ovl);

	mtk_rdma_config(dsi->rdma, timing);
	mtk_rdma_start(dsi->rdma);

	mtk_color_config(dsi->color, timing);
	mtk_color_start(dsi->color);

	mtk_ccorr_config(dsi->ccorr, timing);
	mtk_ccorr_start(dsi->ccorr);

	mtk_aal_config(dsi->aal, timing);
	mtk_aal_start(dsi->aal);

	mtk_gamma_config(dsi->gamma, timing);
	mtk_gamma_start(dsi->gamma);

	mtk_dither_config(dsi->dither, timing);
	mtk_dither_start(dsi->dither);

	ret = mtk_dsi_power_on(dsi, timing);
	if (ret) {
		dev_err(dev, "failed to power on\n");
		return ret;
	}

	mtk_dsi_set_display_mode(dsi, timing);

	mtk_ovl_config(dsi->ovl, timing);

	mtk_ovl_layer(dsi->ovl, plat, timing);

	mmu_set_region_dcache_behaviour(plat->base, ALIGN(plat->size, MMU_SECTION_SIZE),
					DCACHE_WRITEBACK);
	video_set_flush_dcache(dev, true);

	ret = panel_set_backlight(dsi->panel, BACKLIGHT_DEFAULT);
	if (ret) {
		dev_err(dev, "failed to set backlight\n");
		return ret;
	}

	ret = panel_enable_backlight(dsi->panel);
	if (ret) {
		dev_err(dev, "failed to enable backlight\n");
		return ret;
	}

	mtk_dsi_output_enable(dsi);

	return 0;
}

static int mtk_dsi_bind(struct udevice *dev)
{
	struct video_uc_plat *plat = dev_get_uclass_plat(dev);

	plat->size = MTK_DSI_MAX_WIDTH * MTK_DSI_MAX_HEIGHT * VNBYTES(VIDEO_BPP32);

	return 0;
}

static const struct udevice_id mtk_dsi_ids[] = {
	{ .compatible = "mediatek,mt8365-dsi" },
	{}
};

U_BOOT_DRIVER(mtk_dsi) = {
	.name	   = "mtk-dsi",
	.id	   = UCLASS_VIDEO,
	.of_match  = mtk_dsi_ids,
	.probe	   = mtk_dsi_probe,
	.bind	   = mtk_dsi_bind,
	.priv_auto = sizeof(struct mtk_dsi_priv),
	.flags	   = DM_FLAG_PRE_RELOC | DM_FLAG_LEAVE_PD_ON,
};
